#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author: Paul Ruth (pruth@renci.org)

import os
import traceback
import re

import functools

import importlib.resources as pkg_resources
from typing import List

import ipywidgets as widgets
from ipywidgets import HTML, Layout
from ipyleaflet import (
    Map,
    Marker,
    Rectangle,
    WidgetControl,
    basemaps,
    Icon,
    LayerGroup,
    ZoomControl,
    AntPath,
    DrawControl,
    CircleMarker,
    FullScreenControl
)
from fabrictestbed.slice_editor import ExperimentTopology, Capacities, ComponentType, ComponentModelType, ServiceType, ComponentCatalog
from fabrictestbed.slice_editor import (
    ExperimentTopology,
    Capacities
)
from fabrictestbed.slice_manager import SliceManager, Status, SliceState

from .abc_topology_editor import AbcTopologyEditor

from .. import images


class GeoTopologyEditor(AbcTopologyEditor):
    # Constants
    # U.S. Default Map Center
    DEFAULT_US_MAP_CENTER = (38.12480976137421, -85.7129)
    DEFAULT_US_MAP_ZOOM = 4.0
    DEFAULT_US_MAP_WIDTH = '100%'
    DEFAULT_US_MAP_HEIGHT = '650px'

    # FABRIC design elements https://fabric-testbed.net/branding/style/
    FABRIC_PRIMARY = '#27aae1'
    FABRIC_PRIMARY_LIGHT = '#cde4ef'
    FABRIC_PRIMARY_DARK = '#078ac1'
    FABRIC_SECONDARY = '#f26522'
    FABRIC_SECONDARY_LIGHT = '#ff8542'
    FABRIC_SECONDARY_DARK = '#d24502'
    FABRIC_BLACK = '#231f20'
    FABRIC_DARK = '#433f40'
    FABRIC_GREY = '#666677'
    FABRIC_LIGHT = '#f3f3f9'
    FABRIC_WHITE = '#ffffff'
    FABRIC_LOGO = "fabric_logo.png"
    SLICE_RACK_IMAGE = "slice_rack.png"
    SERVER_IMAGE = "server.png"

    # Default Widget values
    DEFAULT_SLICE_NAME_VALUE = 'New_Slice'

    DEFAULT_NODE_SELECT_VALUE = '<Choose Node>'
    DEFAULT_NODE_SITE_VALUE = '<Choose Site>'
    DEFAULT_NODE_CORE_VALUE = 1
    DEFAULT_NODE_RAM_VALUE = 6
    DEFAULT_NODE_DISK_VALUE = 4
    DEFAULT_NODE_IMAGE_VALUE = '<Choose Image>'
    DEFAULT_NODE_IMAGE_OPTIONS = [DEFAULT_NODE_IMAGE_VALUE, 'default_ubuntu_20', 'default_centos_8']
    DEFAULT_NODE_IMAGE_TYPE_VALUE = 'qcow2'
    DEFAULT_NODE_IMAGE_TYPE_OPTIONS = [DEFAULT_NODE_IMAGE_TYPE_VALUE]

    DEFAULT_DASHBOARD = 'slice_dashboard'
    #Current node selected in node dashboard
    current_node = None

    DEFAULT_COMPONENT_MODEL=ComponentModelType.SharedNIC_ConnectX_6


    def __init__(self):
        """
        Constructor
        :return:
        """
        super().__init__()

        self.canvas = None

        self.available_resources_layer_group = LayerGroup(layers=())
        self.widget_layer_group = LayerGroup(layers=())
        self.current_slice_layer_group = LayerGroup(layers=())
        self.slice_layer_groups = LayerGroup(layers=())
        self.dashboards = None
        self.dashboards_buttons = {}

        self.base_overflow_y = 'hidden'
        self.base_min_height = '30px'
        self.base_layout = Layout(min_height='30px', overflow_y='hidden')

        # VBox for dashbords
        self.editor_dashboard = None
        # VBox for control panel
        self.control_panel = None

        print('Creating VisualSliceEditor')

        # Create the Canvas
        self.canvas = Map(basemap=basemaps.Esri.WorldStreetMap,
                          center=self.DEFAULT_US_MAP_CENTER,
                          zoom=self.DEFAULT_US_MAP_ZOOM,
                          zoom_control=False,
                          layout=Layout(width=self.DEFAULT_US_MAP_WIDTH,
                                        height=self.DEFAULT_US_MAP_HEIGHT))

        # LayerGroup for all sites remove_layer
        #self.available_resources_layer_group = LayerGroup(layers=())
        # LayerGroup for all widgets and interface
        #self.widget_layer_group = LayerGroup(layers=())
        # List of a LayerGroups, one for each slice (one for now)
        #self.current_slice_layer_group = LayerGroup(layers=())
        #self.slice_layer_groups = [self.current_slice_layer_group]

        # Add LayerGroups to Canvas
        self.canvas.add_layer(self.available_resources_layer_group)
        self.canvas.add_layer(self.widget_layer_group)
        self.canvas.add_layer(self.current_slice_layer_group)

        #logo_file_path = pkg_resources.read_binary(images, self.FABRIC_LOGO)
        #file = open(logo_file_path, "rb")
        #image = file.read()
        image = pkg_resources.read_binary(images, self.FABRIC_LOGO)
        title_widget = widgets.Image(
            value=image,
            format='png',
            width=300,
            # height=400,
        )

        title_control = WidgetControl(widget=title_widget, position='topleft')
        self.widget_layer_group.add_layer(title_widget)
        self.canvas.add_control(title_control)

        # INIALIZE Editing Dashboard
        self.editor_dashboard = widgets.VBox(layout=Layout(height='535px', width='250px', padding="4px",
                                                           overflow_y='scroll', flex_flow='column', display='flex'))

        editor_dashboard_control = WidgetControl(widget=self.editor_dashboard, position='topright')
        self.canvas.add_control(editor_dashboard_control)

        self.dashboards = {'slice_dashboard': {}, 'link_dashboard': {}}

        self.dashboards['slice_dashboard']['widget_list'] = self.init_slice_dashboard()
        self.init_node_dashboard()
        self.dashboards['link_dashboard']['widget_list'] = self.init_link_dashboard()

        self.current_dashboard = self.dashboards[self.DEFAULT_DASHBOARD]

        # Set defaul dashboard to slice_dashboard
        self.editor_dashboard.children = self.dashboards['slice_dashboard']['widget_list']

        # Initialize control panel
        self.control_panel = widgets.VBox(layout=Layout(height='70px', width='250px', overflow_y='auto',
                                                        padding="4px"))
        control_panel_control = WidgetControl(widget=self.control_panel, position='bottomright')
        self.canvas.add_control(control_panel_control)


        #  LINK Dashboard BTN
        edit_link_btn = widgets.Button(
            description='Links',
            disabled=False,
            tooltip='Click to edit link',
            layout=Layout(width='120px'))
        edit_link_btn.style.button_color = self.FABRIC_PRIMARY
        edit_link_btn.on_click(self.set_edit_link_dashboard)
        self.dashboards['link_dashboard']['button'] = edit_link_btn

        edit_link_control = WidgetControl(widget=edit_link_btn, position='bottomright')


        #  Node Dashboard BTN
        edit_node_btn = widgets.Button(
            description='Nodes',
            disabled=False,
            tooltip='Click to edit node',
            layout=Layout(width='120px'))
        edit_node_btn.style.button_color = self.FABRIC_PRIMARY
        edit_node_btn.on_click(self.set_edit_node_dashboard)
        self.dashboards['node_dashboard']['button'] = edit_node_btn
        edit_node_control = WidgetControl(widget=edit_node_btn, position='bottomright')


        # Slice Dashboard BTN
        edit_slice_btn = widgets.Button(
            description='Slice',
            disabled=False,
            tooltip='Click to edit slice',
            layout=Layout(width='120px'))
        edit_slice_btn.style.button_color = self.FABRIC_PRIMARY
        edit_slice_btn.on_click(self.set_edit_slice_dashboard)
        self.dashboards['slice_dashboard']['button'] = edit_slice_btn
        edit_slice_control = WidgetControl(widget=edit_slice_btn, position='bottomright')

        control_panel_1_hbox = HTML('<center><b>Select Dashboard</b></center>')
        control_panel_2_hbox = widgets.HBox([edit_slice_btn, edit_node_btn, edit_link_btn])
        self.control_panel.children = [control_panel_1_hbox,
                                       control_panel_2_hbox,
                                       ]

        # Set the current dashboard
        self.set_dashboard(self.DEFAULT_DASHBOARD)

        # Add General Canvas Controls
        self.canvas.add_control(FullScreenControl(position='bottomleft'))

        zoom_control = ZoomControl(position='bottomleft')
        self.canvas.observe(self.zoom_control, names='zoom')
        self.canvas.add_control(zoom_control)
        print(zoom_control)

        # XXXXXXX EXPERIMENTAL: Draw Control for shapes on Canvas
        # DRAW CONTROL - MAYBE USE TO DRAW LINKS?
        dc = DrawControl(circlemarker={}, polygon={}, position='bottomleft')
        print(str(dc))
        feature_collection = {
            'type': 'FeatureCollection',
            'features': []}
        dc.on_draw(self.handle_draw)
        self.canvas.add_control(dc)

    def  set_dashboard(self, dashboard_name):
        #Unset the old dashbord
        self.current_dashboard['button'].style.button_color = self.FABRIC_PRIMARY

        #Set the new dashboard
        self.current_dashboard = self.dashboards[dashboard_name]
        self.current_dashboard['button'].style.button_color = self.FABRIC_SECONDARY



    def click_site(self, site_name, **kwargs):
        """
        Handle click on a site
        :param site_name: site name
        :param kwargs:
        :return:
        """
        print("click_site: " + str(kwargs))
        print("site: " + str(site_name))
        event = kwargs['event']
        coordinates = kwargs['coordinates']

        #Toggle Site detail for current experiment
        if self.site_detail == False:
            self.site_detail = True
        else:
            self.site_detail = False


        self.redraw_map()

    def dbclick_site(self, site_name, **kwargs):
        """
        Handle double click on a site
        :param site_name: site name
        :param kwargs:
        :return:
        """
        print("dbclick_site: " + str(kwargs))
        print("site: " + str(site_name))
        event = kwargs['event']
        # event.stopImmediatePropagation()

    def mouseover_site(self, site_name, **kwargs):
        """
        Handle mouse over a site
        :param site_name: site name
        :param kwargs:
        :return:
        """
        print("mouseover_site")
        print("site: " + str(site_name))
        print("double_click_zoom: " + str(self.canvas.double_click_zoom))
        self.canvas.disableDoubleClickZoom = True
        # self.canvas.double_click_zoom = False;

    def mouseout_site(self, site_name, **kwargs):
        """
        Handle mouse out on a site
        :param site_name: site name
        :param kwargs:
        :return:
        """
        print("mouseout_site")
        print("site: " + str(site_name))
        print("double_click_zoom: " + str(self.canvas.double_click_zoom))
        self.canvas.disableDoubleClickZoom = False

    @staticmethod
    def ant_path(path_name, **kwargs):
        """
        Print ant paths
        :param path_name: path name
        :param kwargs:
        :return:
        """
        print("ant_path")
        print("path_name: " + str(path_name))
        print(kwargs)

    def update_sites(self):
        """
        Update Sites
        :return:
        """
        print('Update Sites')
        # Query FABRIC for updated resources
        self.pull_advertised_topology()

        # Clear the resource layers
        self.available_resources_layer_group.clear_layers()

        # Add the sites
        site_name_list = []
        for site in self.advertised_topology.sites.values():
            try:
                site_marker = CircleMarker()
                site_marker.location = site.get_property("location").to_latlon()
                site_marker.radius = 10
                site_marker.color = self.FABRIC_PRIMARY
                site_marker.fill_color = self.FABRIC_PRIMARY

                self.available_resources_layer_group.add_layer(site_marker)

                site_marker.on_click(functools.partial(self.click_site, site_name=site.get_property("name")))
                site_marker.on_dblclick(functools.partial(self.dbclick_site, site_name=site.get_property("name")))
                site_marker.on_mouseover(functools.partial(self.mouseover_site, site_name=site.get_property("name")))
                site_marker.on_mouseout(functools.partial(self.mouseout_site, site_name=site.get_property("name")))
                site_name_list.append(site.get_property("name"))

            except Exception as e:
                print('Failed to add site: ' + str(site) + '. Error: ' + str(e))
                # traceback.print_exc()

        for link_name, link in self.advertised_topology.links.items():
            try:
                #Source
                source_interface = link.interface_list[0]
                source_parent = self.advertised_topology.get_parent_element(source_interface)
                source_node=self.advertised_topology.get_owner_node(source_parent)
                print ("Source node: {}".format(source_node))
                source_location=source_node.get_property("location").to_latlon()

                #Target
                target_interface = link.interface_list[1]
                target_parent = self.advertised_topology.get_parent_element(target_interface)
                target_node=self.advertised_topology.get_owner_node(target_parent)
                print ("Target node: {}".format(target_node))
                target_location=target_node.get_property("location").to_latlon()


                #Build edge

                ant_path = AntPath(
                    locations=[source_location,target_location],
                    dash_array=[1, 10],
                    delay=1000,
                    color='#7590ba',
                    pulse_color=self.FABRIC_PRIMARY,
                    paused=True,
                    hardwareAccelerated=True,
                    description='Task'
                )

                ant_path.on_click(functools.partial(self.ant_path, path_name=link_name))
                self.available_resources_layer_group.add_layer(ant_path)
            except:
                print("Skiping link: {}".format(link) )




        #
        # sites = []
        # site_locations = []
        # for site_name, site in self.advertised_topology.sites.items():
        #     try:
        #         site_locations.append(site.get_property("location").to_latlon())
        #         sites.append(site)
        #     except Exception as e:
        #         print('Failed to add site: ' + str(site) + '. Error: ' + str(e))
        #         # traceback.print_exc()
        #
        # paths = {}
        # for i, site in enumerate(sites):
        #     # site_locations.append(site.get_property("location").to_latlon())
        #     path_locations = [sites[i].get_property("location").to_latlon(),
        #                       sites[(i + 1) % len(sites)].get_property("location").to_latlon()]
        #     ant_path = AntPath(
        #         locations=path_locations,
        #         dash_array=[1, 10],
        #         delay=1000,
        #         color='#7590ba',
        #         pulse_color=self.FABRIC_PRIMARY,
        #         paused=True,
        #         hardwareAccelerated=True,
        #         description='Task'
        #     )
        #
        #     ant_path.on_click(functools.partial(self.ant_path, path_name=i))
        #     self.available_resources_layer_group.add_layer(ant_path)

        # Update node Dashboard
        site_name_list = ['<Choose Site>'] + sorted(site_name_list)
        site_name_widget = widgets.Dropdown(
            # placeholder='Enter Site Name',
            options=site_name_list,
            value='<Choose Site>',
            disabled=False,
            ensure_option=True,
            tooltip='Enter Site Name',
        )
        self.dashboards['node_dashboard']['site_name_widget'] = site_name_widget
        self.dashboards['node_dashboard']['site_name_hbox'] = widgets.HBox([widgets.Label(value="Site: ",
                                                                                          layout=Layout(width='70px',
                                                                                                        min_height=self.base_min_height,
                                                                                                        overflow_y=self.base_overflow_y)),
                                                                            site_name_widget], layout=self.base_layout)
        self.build_node_dashboard_widget_list()

    def init_node_dashboard(self):
        """
        Initialize Node dashboard
        :return:
        """
        dashboard = self.dashboards['node_dashboard'] = {}

        # Init a dictionary of compoenents
        dashboard['component_widgets'] = []
        dashboard['widget_list'] = []

        # Create Node Propertiees dashboard
        header = HTML('<center><b>Node Dashboard</b></center>')
        dashboard['header'] = header

        select_node_widget = widgets.Select(
            options=[ self.DEFAULT_NODE_SELECT_VALUE ],
            value=self.DEFAULT_NODE_SELECT_VALUE ,
            disabled=False,
            ensure_option=True,
            tooltip='Choose node to edit',
            layout=self.base_layout
        )
        select_node_widget.observe(self.node_dashboard_select_node_event, names='value')
        dashboard['select_node_widget'] = select_node_widget
        select_node_hbox = widgets.HBox([widgets.Label(value="Select Node: ",
                                         layout=Layout(width='150px',
                                                       min_height=self.base_min_height,
                                                       overflow_y=self.base_overflow_y)),
                                         select_node_widget],
                                         layout=self.base_layout)
        dashboard['select_node_hbox'] = select_node_hbox

        # Edit Node
        # edit_header = HTML('<center><b>Edit Node</b></center>')
        edit_header = HTML('<center><b><hr></b><b>Edit Node</b></center>')
        dashboard['edit_header'] = edit_header

        # Edit node fields
        node_name_widget = widgets.Text(
            # value='',
            placeholder='Enter New Node Name',
            disabled=False,
            tooltip='Enter Node Name',
            layout=self.base_layout
        )
        # node_name_widget.observe(self.node_dashboard_node_name_event, names='value')
        dashboard['node_name_widget'] = node_name_widget
        node_name_hbox = widgets.HBox([widgets.Label(value="Name: ", layout=Layout(width='70px',
                                                                                   min_height=self.base_min_height,
                                                                                   overflow_y=self.base_overflow_y)),
                                       node_name_widget], layout=self.base_layout)
        dashboard['node_name_hbox'] = node_name_hbox

        site_name_widget = widgets.Dropdown(
            # placeholder='Enter Site Name',
            options=[self.DEFAULT_NODE_SITE_VALUE],
            value=self.DEFAULT_NODE_SITE_VALUE,
            disabled=False,
            ensure_option=True,
            tooltip='Enter Site Name',
            layout=self.base_layout
        )
        dashboard['site_name_widget'] = site_name_widget
        site_name_hbox = widgets.HBox([widgets.Label(value="Site: ", layout=Layout(width='70px',
                                                                                   min_height=self.base_min_height,
                                                                                   overflow_y=self.base_overflow_y)),
                                       site_name_widget], layout=self.base_layout)
        dashboard['site_name_hbox'] = site_name_hbox

        core_slider = widgets.IntSlider(min=2, max=32, step=2, value=self.DEFAULT_NODE_CORE_VALUE)
        core_slider_hbox = widgets.HBox([widgets.Label(value="Cores: ", layout=Layout(width='100px',
                                                                                      min_height=self.base_min_height,
                                                                                      overflow_y=self.base_overflow_y)),
                                         core_slider], layout=self.base_layout)
        dashboard['core_slider'] = core_slider
        dashboard['core_slider_hbox'] = core_slider_hbox

        ram_slider = widgets.IntSlider(min=2, max=64, step=2, value=self.DEFAULT_NODE_RAM_VALUE)
        ram_slider_hbox = widgets.HBox([widgets.Label(value="RAM (G): ", layout=Layout(width='100px',
                                                                                       min_height=self.base_min_height,
                                                                                       overflow_y=self.base_overflow_y)),
                                        ram_slider], layout=self.base_layout)
        dashboard['ram_slider'] = ram_slider
        dashboard['ram_slider_hbox'] = ram_slider_hbox

        disk_slider = widgets.FloatSlider(value=self.DEFAULT_NODE_DISK_VALUE, min=2, max=130, step=0.1,
                                          readout_format='1.0f')
        disk_slider_hbox = widgets.HBox([widgets.Label(value="Disk (G): ", layout=Layout(width='100px',
                                                                                         min_height=self.base_min_height,
                                                                                         overflow_y=self.base_overflow_y)),
                                         disk_slider], layout=self.base_layout)
        dashboard['disk_slider'] = disk_slider
        dashboard['disk_slider_hbox'] = disk_slider_hbox

        image_type_widget = widgets.Dropdown(
            # placeholder='Enter Site Name',
            options=self.DEFAULT_NODE_IMAGE_TYPE_OPTIONS,
            value=self.DEFAULT_NODE_IMAGE_TYPE_VALUE,
            disabled=False,
            ensure_option=True,
            tooltip='Choose Image Type',
            layout=self.base_layout
        )
        dashboard['image_type_widget'] = image_type_widget
        image_type_widget_hbox = widgets.HBox([widgets.Label(value="Image Type: ", layout=Layout(width='150px',
                                                                                                 min_height=self.base_min_height,
                                                                                                 overflow_y=self.base_overflow_y)),
                                               image_type_widget], layout=self.base_layout)
        dashboard['image_type_widget_hbox'] = image_type_widget_hbox

        image_widget = widgets.Dropdown(
            options=self.DEFAULT_NODE_IMAGE_OPTIONS,
            value=self.DEFAULT_NODE_IMAGE_VALUE,
            tooltip='Choose Image',
            ensure_option=True,
            disabled=False,
            layout=self.base_layout
        )
        dashboard['image_widget'] = image_widget
        image_widget_hbox = widgets.HBox([widgets.Label(value="Image: ", layout=Layout(width='150px',
                                                                                       min_height=self.base_min_height,
                                                                                       overflow_y=self.base_overflow_y)),
                                          image_widget], layout=self.base_layout)
        dashboard['image_widget_hbox'] = image_widget_hbox

        # Trying to get Accordion to work
        # edit_node_vbox = widgets.VBox([node_name_hbox,
        #                                site_name_hbox,
        #                                core_slider_hbox ,
        #                                ram_slider_hbox ,
        #                                disk_slider_hbox ,
        #                                image_widget_hbox,], layout=Layout(width='100%', min_height='300px', overflow_y='hidden'))
        # edit_node_accordion = widgets.Accordion(children=[edit_node_vbox ], layout=Layout(width='100%', min_height='300px', overflow_y='unset'))
        # edit_node_accordion.set_title(0, 'Edit Node Properties')


        delete_node_btn = widgets.Button(
            description='Delete',
            disabled=False,
            tooltip='click to delete node',
            layout=self.base_layout
        )
        dashboard['delete_node_btn'] = delete_node_btn
        delete_node_btn.style.button_color = self.FABRIC_PRIMARY
        delete_node_btn.on_click(self.node_dashboard_delete_node)

        add_node_btn = widgets.Button(
            description='Add',
            disabled=False,
            tooltip='click to add a new node',
            layout=self.base_layout
        )
        dashboard['add_node_btn'] = add_node_btn
        add_node_btn.style.button_color = self.FABRIC_PRIMARY
        add_node_btn.on_click(self.node_dashboard_add_node)

        node_edit_button_hbox = widgets.HBox([add_node_btn, delete_node_btn], layout=self.base_layout)
        dashboard['node_edit_button_hbox'] = node_edit_button_hbox

        component_header = HTML('<center><b><hr></b><b>Components</b></center>')
        dashboard['component_header'] = component_header

        add_component_btn = widgets.Button(
            description='Add Component',
            disabled=False,
            tooltip='click to add a component',
            layout=self.base_layout
        )
        dashboard['add_component_btn'] = add_component_btn
        add_component_btn.style.button_color = self.FABRIC_PRIMARY
        add_component_btn.on_click(self.node_dashboard_add_component)
        add_component_button_hbox = widgets.HBox([add_component_btn], layout=self.base_layout)
        dashboard['add_component_button_hbox'] = add_component_button_hbox

        #Initi component dict
        dashboard['component_widgets'] = {}

        self.build_node_dashboard_widget_list()

    def build_node_dashboard_widget_list(self):
        """
        Build node dashboard widget list
        :return:
        """
        dashboard = self.dashboards['node_dashboard']

        node_dashboard_list = [dashboard['header'],
                               dashboard['select_node_hbox'],
                               dashboard['node_edit_button_hbox']
                               ]

        if dashboard['select_node_widget'].value != '<Choose Node>':
            node_dashboard_list += [dashboard['edit_header'],
                                    dashboard['node_name_hbox'],
                                    dashboard['site_name_hbox'],
                                    dashboard['core_slider_hbox'],
                                    dashboard['ram_slider_hbox'],
                                    dashboard['disk_slider_hbox'],
                                    dashboard['image_widget_hbox'],
                                    dashboard['component_header'],
                                    dashboard['add_component_button_hbox'],
                                    ]
            for component_name, component in dashboard['component_widgets'].items():
                print(component)
                node_dashboard_list.append(component['component_header'])
                node_dashboard_list.append(component['component_name_hbox'])
                #node_dashboard_list.append(component['component_type_hbox'])
                node_dashboard_list.append(component['component_model_hbox'])

        dashboard['widget_list'] = node_dashboard_list

        return node_dashboard_list

    def init_slice_dashboard(self):
        """
        Initialize slice dashboard
        :return:
        """
        # Create Slice Propertiees dashboard
        dashboard = self.dashboards['slice_dashboard']

        header = HTML('<center><b>Slice Dashboard</b></center>')

        slice_select_widget = widgets.Select(
            options=[self.DEFAULT_SLICE_SELECT_VALUE],
            value=self.DEFAULT_SLICE_SELECT_VALUE,
            tooltip='Choose Slice',
            ensure_option=True,
            disabled=False,
            layout=self.base_layout
        )
        slice_select_widget.observe(self.slice_dashboard_select_slice, names='value')
        dashboard['slice_select_widget'] = slice_select_widget
        slice_select_widget_hbox = widgets.HBox([widgets.Label(value="Current Slice: ",
                                                               layout=Layout(width='150px',
                                                                             min_height=self.base_min_height,
                                                                             overflow_y=self.base_overflow_y)),
                                                 slice_select_widget], layout=self.base_layout)
        dashboard['slice_select_widget_hbox'] = slice_select_widget_hbox

        slice_name_widget = widgets.Text(
            # value='<New Slice Name>',
            placeholder='Enter New Slice Name',
            disabled=False,
            tooltip='Enter Slice Name',
            # layout=Layout(width='170px')
        )
        dashboard['slice_name_widget'] = slice_name_widget
        slice_name_hbox = widgets.HBox([widgets.Label(value="Name: ", layout=Layout(width='70px')), slice_name_widget])

        # ADD NODE BTN
        new_slice_btn = widgets.Button(
            description='New Slice',
            disabled=False,
            tooltip='click to create a new slice',
        )
        new_slice_btn.style.button_color = self.FABRIC_PRIMARY
        new_slice_btn.on_click(self.slice_dashboard_new_slice)

        update_slices_btn = widgets.Button(
            description='Update',
            disabled=False,
            tooltip='click get a the current list of slices',
        )
        update_slices_btn.style.button_color = self.FABRIC_PRIMARY
        update_slices_btn.on_click(self.slice_dashboard_update_slices)

        # open_slice_btn = widgets.Button(
        #    description='Open Slice',
        #    disabled=False,
        #    tooltip='Click to submit the slice',
        # )
        # open_slice_btn.style.button_color = self.FABRIC_PRIMARY
        # open_slice_btn.on_click(self.slice_dashboard_open_slice)

        submit_slice_btn = widgets.Button(
            description='Submit Slice',
            disabled=False,
            tooltip='Click to submit the slice',
        )
        submit_slice_btn.style.button_color = self.FABRIC_PRIMARY
        submit_slice_btn.on_click(self.slice_dashboard_submit_slice)

        delete_experiment_btn = widgets.Button(
            description='Delete Slice',
            disabled=False,
            tooltip='click to delete the slice',
        )
        delete_experiment_btn.style.button_color = self.FABRIC_PRIMARY
        delete_experiment_btn.on_click(self.slice_dashboard_delete_experiment)
        button_1_hbox = widgets.HBox([new_slice_btn, update_slices_btn])
        button_2_hbox = widgets.HBox([submit_slice_btn, delete_experiment_btn])

        # Slice Status Text
        slice_status_header = HTML('<center><b>Slice Status</b></center>')

        slice_experiment_state_label = widgets.Label(value="Expeirment State: ",
                                                      layout=Layout(width='80px',
                                                                    min_height=self.base_min_height,
                                                                    overflow_y=self.base_overflow_y))
        slice_experiment_state_value = widgets.Label(value="", layout=Layout(width='150px',
                                                                              min_height=self.base_min_height,
                                                                              overflow_y=self.base_overflow_y))
        dashboard['slice_experiment_state_value'] = slice_experiment_state_value
        slice_experiment_state_widget_hbox = widgets.HBox([slice_experiment_state_label,
                                                            slice_experiment_state_value])

        slice_status_slice_name_label = widgets.Label(value="Slice Name: ",
                                                      layout=Layout(width='80px',
                                                                    min_height=self.base_min_height,
                                                                    overflow_y=self.base_overflow_y))
        slice_status_slice_name_value = widgets.Label(value="", layout=Layout(width='150px',
                                                                              min_height=self.base_min_height,
                                                                              overflow_y=self.base_overflow_y))
        dashboard['slice_status_slice_name_value'] = slice_status_slice_name_value
        slice_status_slice_name_widget_hbox = widgets.HBox([slice_status_slice_name_label,
                                                            slice_status_slice_name_value])

        slice_status_slice_id_label = widgets.Label(value="Slice ID: ", layout=Layout(width='80px',
                                                                                      min_height=self.base_min_height,
                                                                                      overflow_y=self.base_overflow_y))
        slice_status_slice_id_value = widgets.Label(value="", layout=Layout(width='150px',
                                                                            min_height=self.base_min_height,
                                                                            overflow_y=self.base_overflow_y))
        dashboard['slice_status_slice_id_value'] = slice_status_slice_id_value
        slice_status_slice_id_widget_hbox = widgets.HBox([slice_status_slice_id_label, slice_status_slice_id_value])

        slice_status_slice_state_label = widgets.Label(value="Slice State: ",
                                                       layout=Layout(width='80px',
                                                                     min_height=self.base_min_height,
                                                                     overflow_y=self.base_overflow_y))
        slice_status_slice_state_value = widgets.Label(value="", layout=Layout(width='150px',
                                                                               min_height=self.base_min_height,
                                                                               overflow_y=self.base_overflow_y))
        dashboard['slice_status_slice_state_value'] = slice_status_slice_state_value
        slice_status_slice_state_widget_hbox = widgets.HBox([slice_status_slice_state_label,
                                                             slice_status_slice_state_value])

        slice_status_lease_end_label = widgets.Label(value="Lease End: ",
                                                     layout=Layout(width='80px',
                                                                   min_height=self.base_min_height,
                                                                   overflow_y=self.base_overflow_y))
        slice_status_lease_end_value = widgets.Label(value="", layout=Layout(width='150px',
                                                                             min_height=self.base_min_height,
                                                                             overflow_y=self.base_overflow_y))
        dashboard['slice_status_lease_end_value'] = slice_status_lease_end_value
        slice_status_lease_end_widget_hbox = widgets.HBox([slice_status_lease_end_label, slice_status_lease_end_value])

        slice_status_graph_id_label = widgets.Label(value="Graph ID: ",
                                                    layout=Layout(width='80px',
                                                                  min_height=self.base_min_height,
                                                                  overflow_y=self.base_overflow_y))
        slice_status_graph_id_value = widgets.Label(value="",
                                                    layout=Layout(width='150px',
                                                                  min_height=self.base_min_height,
                                                                  overflow_y=self.base_overflow_y))

        dashboard['slice_status_graph_id_value'] = slice_status_graph_id_value
        slice_status_graph_id_widget_hbox = widgets.HBox([slice_status_graph_id_label, slice_status_graph_id_value])

        slice_status_vbox = widgets.VBox([slice_status_header,
                                          slice_experiment_state_widget_hbox,
                                          slice_status_slice_name_widget_hbox,
                                          slice_status_slice_id_widget_hbox,
                                          slice_status_slice_state_widget_hbox,
                                          slice_status_lease_end_widget_hbox,
                                          slice_status_graph_id_widget_hbox])

        return [header,
                slice_select_widget_hbox,
                slice_name_hbox,
                button_1_hbox,
                button_2_hbox,
                slice_status_vbox]

    def init_link_dashboard(self):
        """
        Initialize the link dashboard
        :return:
        """
        dashboard = self.dashboards['link_dashboard']

        # Create link Propertiees dashboard
        header = HTML('<center><b>Link Dashboard</b></center>')
        link_name_widget = widgets.Text(
            value='',
            placeholder='Enter Link Name',
            disabled=False,
            tooltip='Enter Link Name',
            # layout=Layout(width='170px')
        )
        dashboard['link_name_widget'] = link_name_widget
        link_name_hbox = widgets.HBox([widgets.Label(value="Name: ", layout=Layout(width='70px')), link_name_widget])
        return [header, link_name_hbox]

    def zoom_control(self, change):
        """
        Zoom control
        :param change:
        :return:
        """
        print("CHANGE: " + str(change))
        print("CANVAS: " + str(self.canvas))
        print("ZOOM: " + str(self.canvas.zoom))
        self.redraw_map()

    def redraw_map(self):
        """
        Redraw slice map
        :return:
        """
        print("redraw_map")

        # Clear slice
        try:
            self.canvas.remove_layer(self.current_slice_layer_group)
        except Exception as e:
            print("layer group not in canvas")
        self.current_slice_layer_group.clear_layers()

        if self.current_experiment is None or 'topology' not in self.current_experiment.keys():
            print("Can not redraw map. No current topology to draw")
            return

        topology = self.current_experiment['topology']

        # slivers = self.slice_manager.slivers(slices=[self.current_experiment['slice']])
        if self.site_detail:
           #print("topo nodes: " + str(self.current_experiment['topology'].nodes))

            slice_rack_markers = []
            count = 0
            #Rectangle scale
            baselength=float(45.0)
            baseheight=float(45.0)
            length=baselength/(pow(2,self.canvas.zoom-1))
            height=baseheight/(pow(2,self.canvas.zoom-1))
            length_offset=length/float(2.0)
            height_offset=height/float(2.0)
            icon_base = 10
            icon_offset = icon_base/(pow(2,float(self.canvas.zoom-1)))

            #print("height_offset: " + str(height_offset) + ", length_offset: " + str(length_offset))
            for node in self.current_experiment['topology'].nodes:
                #print("Node: " + str(self.current_experiment['topology'].nodes[node]))
                site = self.current_experiment['topology'].nodes[node].get_property(pname='site')
                #print("sites: " + str(self.advertised_topology.sites))
                #printSite: " +  site)

                location = self.advertised_topology.sites[site].get_property(pname='location').to_latlon()
                if site not in slice_rack_markers:

                    #print("CREATING RECTANGLE " + str(location[0]) + ", " + str(location[1]))

                    self.rectangle = Rectangle(bounds=((location[0]-height_offset,location[1]-length_offset),(location[0]+height_offset,location[1]+length_offset)),
                    #self.rectangle = Rectangle(bounds=((location[0]+2,location[1]-4),(location[0]-5,location[1]+4)),
                                          color=self.FABRIC_PRIMARY,
                                          fill_color=self.FABRIC_LIGHT,
                                          fill_opacity=0.6)
                    self.current_slice_layer_group.add_layer(self.rectangle)
                    slice_rack_markers.append(site)

                #print("ADDING MARKER: " +  site)



                icon = Icon(icon_url=self.SERVER_IMAGE,icon_size=[50, 20])
                mark = Marker(location=self.location_plus(location,plus_x=0,plus_y=(0-count)*icon_offset), icon=icon)
                count += 1
                self.current_slice_layer_group.add_layer(mark)
        else:
            slice_rack_markers = []
            for node in self.current_experiment['topology'].nodes:
                #print("Node: " + str(self.current_experiment['topology'].nodes[node]))
                site = self.current_experiment['topology'].nodes[node].get_property(pname='site')

                #Skip icon if site for node has not been picked
                if site == self.DEFAULT_NODE_SITE_VALUE:
                    continue

                if site not in slice_rack_markers:
                    #print("ADDING MARKER: " + site)
                    location = self.advertised_topology.sites[site].get_property(pname='location').to_latlon()
                    icon = Icon(icon_url=self.SLICE_RACK_IMAGE)
                    mark = Marker(location=location, icon=icon)
                    slice_rack_markers.append(site)
                    self.current_slice_layer_group.add_layer(mark)

        self.canvas.add_layer(self.current_slice_layer_group)

    @staticmethod
    def location_plus(location, plus_x, plus_y):
        """
        Location plus
        :param location: current location
        :param plus_x:
        :param plus_y:
        :return:
        """
        return location[0] + plus_y, location[1] + plus_x

    def update_edit_slice_dashboard(self):
        """
        Update Edit slice dashboard
        :return:
        """
        self.editor_dashboard.children = self.dashboards['slice_dashboard']['widget_list']

    def set_edit_slice_dashboard(self, button):
        """
        Set Edit slice dashboard
        :param button:
        :return:
        """
        self.save_node_widget_data()
        self.set_dashboard('slice_dashboard')
        self.update_edit_slice_dashboard()

    def rebuild_node_dashboard(self, node_name=None):
        """
        Rebuild Node dashboard
        :param node_name:
        :return:
        """
        print(" XX REBUILD NODE DASHBOARD XX")
        dashboard = self.dashboards['node_dashboard']
        # Init a dictionary of components
        dashboard['component_widgets'] = []

        #Nodes from topology
        experiment_topology = self.current_experiment['topology']
        nodes = experiment_topology.nodes.keys()

        if len(nodes) == 0:
            self.dashboards['node_dashboard']['select_node_widget'].options = [ self.DEFAULT_NODE_SELECT_VALUE ]
            self.dashboards['node_dashboard']['select_node_widget'].value = self.DEFAULT_NODE_SELECT_VALUE

            dashboard['node_name_widget'].value = ''
            dashboard['site_name_widget'].value = self.DEFAULT_NODE_SITE_VALUE
            dashboard['core_slider'].value = self.DEFAULT_NODE_CORE_VALUE
            dashboard['ram_slider'].value = self.DEFAULT_NODE_RAM_VALUE
            dashboard['disk_slider'].value = self.DEFAULT_NODE_DISK_VALUE
            dashboard['image_type_widget'].value = self.DEFAULT_NODE_IMAGE_TYPE_VALUE
            dashboard['image_widget'].value = self.DEFAULT_NODE_IMAGE_VALUE

            self.dashboards['node_dashboard']['component_widgets'] = {}

        else:
            # Update node select widget
            self.dashboards['node_dashboard']['select_node_widget'].options = sorted(self.get_node_name_list())
            if node_name == None:
                node_name = sorted(self.get_node_name_list())[0]

            self.update_select_node_widget_option_name(node_name)
            dashboard['select_node_widget'].observe(self.node_dashboard_select_node_event, names='')
            dashboard['node_name_widget'].value = node_name
            dashboard['select_node_widget'].observe(self.node_dashboard_select_node_event, names='value')

            component_header = HTML('<center><b><hr></b><b>Components</b></center>')
            dashboard['component_header'] = component_header

            add_component_btn = widgets.Button(
                description='Add Component',
                disabled=False,
                tooltip='click to add a component',
                layout=self.base_layout
            )
            dashboard['add_component_btn'] = add_component_btn
            add_component_btn.style.button_color = self.FABRIC_PRIMARY
            add_component_btn.on_click(self.node_dashboard_add_component)
            add_component_button_hbox = widgets.HBox([add_component_btn], layout=self.base_layout)
            dashboard['add_component_button_hbox'] = add_component_button_hbox

            component_widgets = {}
            for component_name, component in self.current_node.components.items():
                component_widget = component_widgets[component_name] = {}

                # Edit Node
                component_header = HTML('<center><b><hr></b></center>')
                component_widget['component_header'] = component_header

                # Edit node fields
                component_name_widget = widgets.Text(
                    value=component.name,
                    placeholder='Enter Component Name',
                    disabled=False,
                    tooltip='Enter Component Name',
                    layout=self.base_layout
                )
                component_widget['component_name_widget'] = component_name_widget
                component_name_hbox = widgets.HBox([widgets.Label(value="Name: ",
                                                                  layout=Layout(width='70px',
                                                                                min_height=self.base_min_height,
                                                                                overflow_y=self.base_overflow_y)),
                                                    component_name_widget], layout=self.base_layout)
                component_widget['component_name_hbox'] = component_name_hbox

                # component_type_widget = widgets.Dropdown(
                #     options=['<Any Type>'] + sorted(self.get_component_type_list()),
                #     value='<Any Type>',
                #     disabled=False,
                #     ensure_option=True,
                #     tooltip='Choose Component Type',
                #     layout=self.base_layout
                # )
                # component_widget['component_type_widget'] = component_type_widget
                # component_type_hbox = widgets.HBox([widgets.Label(value="Type: ",
                #                                                   layout=Layout(width='70px',
                #                                                                 min_height=self.base_min_height,
                #                                                                 overflow_y=self.base_overflow_y)),
                #                                     component_type_widget], layout=self.base_layout)
                # component_widget['component_type_hbox'] = component_type_hbox

                component_model_widget = widgets.Dropdown(
                    options=['<Choose Component Model>'] + self.get_component_widget_options_list(),
                    value='<Choose Component Model>',
                    disabled=False,
                    ensure_option=True,
                    tooltip='Choose Component Model',
                    layout=self.base_layout
                )
                component_widget['component_model_widget'] = component_model_widget
                component_model_hbox = widgets.HBox([widgets.Label(value="Model: ",
                                                                   layout=Layout(width='70px',
                                                                                 min_height=self.base_min_height,
                                                                                 overflow_y=self.base_overflow_y)),
                                                     component_model_widget], layout=self.base_layout)
                component_widget['component_model_hbox'] = component_model_hbox

            self.dashboards['node_dashboard']['component_widgets'] = component_widgets






        self.build_node_dashboard_widget_list()
        self.editor_dashboard.children = self.dashboards['node_dashboard']['widget_list']

    def set_edit_node_dashboard(self, button):
        """
        Set Edit node dashboard
        :param button:
        :return:
        """
        self.set_dashboard('node_dashboard')

        self.rebuild_node_dashboard()

    def update_edit_link_dashboard(self):
        """
        Update Edit Link dashboard
        :return:
        """
        # self.build_link_dashboard_widget_list()
        self.editor_dashboard.children = self.dashboards['link_dashboard']['widget_list']

    def set_edit_link_dashboard(self, button):
        """
        Set edit link dashboard
        :param button:
        :return:
        """
        self.set_dashboard('link_dashboard')

        self.update_edit_link_dashboard()

    def add_node(self, node_name):
        """
        Add node
        :param node_name:
        :return:
        """
        try:
            print("geo.add_node: {}".format(node_name))
            super().add_node(node_name)

            # Update the map markers/connections

            # Update node dashboard values
            self.dashboards['node_dashboard']['node_name_widget'].value = node_name
            self.dashboards['node_dashboard']['site_name_widget'].value = self.DEFAULT_NODE_SITE_VALUE
            self.dashboards['node_dashboard']['core_slider'].value = self.DEFAULT_NODE_CORE_VALUE
            self.dashboards['node_dashboard']['ram_slider'].value = self.DEFAULT_NODE_RAM_VALUE
            self.dashboards['node_dashboard']['disk_slider'].value = self.DEFAULT_NODE_DISK_VALUE
            self.dashboards['node_dashboard']['image_widget'].value = self.DEFAULT_NODE_IMAGE_VALUE
            self.dashboards['node_dashboard']['image_type_widget'].value = self.DEFAULT_NODE_IMAGE_TYPE_VALUE
            # self.update_select_node_widget_option_name(node_name=node_name, new_node=True)
            #self.rebuild_node_dashboard(node_name=node_name)

        except Exception as e:
            # TODO: Should create popup or other user facing error message
            print('Failed to add node. Error: ' + str(e))
            traceback.print_exc()



    def load_node(self, node_name):
        """
        Load a node
        :param node_name: node name
        :return:
        """
        try:
            node = self.current_experiment['topology'].nodes[node_name]
            self.current_node = node


            self.dashboards['node_dashboard']['node_name_widget'].value = node_name
            self.dashboards['node_dashboard']['site_name_widget'].value = node.get_property(pname='site')
            self.dashboards['node_dashboard']['core_slider'].value = int(self.get_capacity_value(node, 'core'))
            self.dashboards['node_dashboard']['ram_slider'].value = int(self.get_capacity_value(node, 'ram'))
            self.dashboards['node_dashboard']['disk_slider'].value = float(self.get_capacity_value(node, 'disk'))
            self.dashboards['node_dashboard']['image_widget'].value = node.get_property(pname='image_ref')
            self.dashboards['node_dashboard']['image_type_widget'].value = node.get_property(pname='image_type')

            #TODO: LOAD Components


            self.current_node = node
        except Exception as e:
            # TODO: Should create popup or other user facing error message
            print('Failed to load node. Error: ' + str(e))
            traceback.print_exc()



    def add_component(self):
        """
        Add component
        :return:
        """
        pass

    def delete_component(self):
        """
        Delete component
        :return:
        """
        pass

    def load_component(self):
        """
        Load component
        :return:
        """
        pass

    def save_component(self):
        """
        Save component
        :return:
        """
        pass

    # Crappy work around for getting capacity values out of topology nodes
    # TODO: FIX THIS
    def get_capacity_value(self, node, name='core'):
        """
        Get capacity value
        :param node:
        :param name:
        :return:
        """
        # name options: 'core', 'ram', 'disk'
        capacities = str(node.get_property(pname='capacities'))

        capacities = capacities.replace('{', '').replace('}', '').replace('G', '').replace(' ', '')
        for line in re.split(",", capacities):
            key_val = re.split(":", line)
            if key_val[0] == name:
                return key_val[1]

    def update_select_node_widget_option_name(self, name):
        """
        Update select node widget option name
        :param name:
        :return:
        """
        self.dashboards['node_dashboard']['select_node_widget'].observe(self.node_dashboard_select_node_event, names='')
        if name:
            self.dashboards['node_dashboard']['select_node_widget'].value = name
        self.dashboards['node_dashboard']['select_node_widget'].observe(self.node_dashboard_select_node_event,
                                                                        names='value')

    def node_dashboard_node_name_event(self, change):
        """
        Node dashboard node name event
        :param change:
        :return:
        """
        print("node_dashboard_node_name_event")
        print(change)

        # Update the name
        old_name = self.dashboards['node_dashboard']['select_node_widget'].value
        new_name = change['new']

        self.update_select_node_widget_option_name(new_name)

    def save_node_widget_data(self):
        self.save_node(topology_node=self.current_node,
                       node_name=self.dashboards['node_dashboard']['node_name_widget'].value,
                       site_name=self.dashboards['node_dashboard']['site_name_widget'].value,
                       cores=self.dashboards['node_dashboard']['core_slider'].value,
                       ram=self.dashboards['node_dashboard']['ram_slider'].value,
                       disk=self.dashboards['node_dashboard']['disk_slider'].value,
                       image=self.dashboards['node_dashboard']['image_widget'].value,
                       image_type=self.dashboards['node_dashboard']['image_type_widget'].value)

    def node_dashboard_select_node_event(self, change):
        """
        Node dashboard select node event
        :param change:
        :return:
        """
        print("node_dashboard_select_node_event")
        print(change)

        # Save old node selection
        old_node_name = change['old']
        if old_node_name != self.DEFAULT_NODE_SELECT_VALUE:
            self.save_node_widget_data()
            #self.save_node(topology_node=self.current_node,
            #               node_name=self.dashboards['node_dashboard']['node_name_widget'].value,
            #               site_name=self.dashboards['node_dashboard']['site_name_widget'].value,
            #               cores=self.dashboards['node_dashboard']['core_slider'].value,
            #               ram=self.dashboards['node_dashboard']['ram_slider'].value,
            #               disk=self.dashboards['node_dashboard']['disk_slider'].value,
            #               image=self.dashboards['node_dashboard']['image_widget'].value,
            #               image_type=self.dashboards['node_dashboard']['image_type_widget'].value)
        # Display new node selection
        new_node_name = change['new']
        if new_node_name != None and new_node_name != self.DEFAULT_NODE_SELECT_VALUE:
            self.load_node(new_node_name)

        # Re-draw node dashboard
        self.rebuild_node_dashboard(new_node_name)

    def node_dashboard_add_node(self, button, **kwargs):
        """
        Node dashboard add node
        :param button:
        :param kwargs:
        :return:
        """
        print('Add Node')

        if self.current_node:
            self.save_node(topology_node=self.current_node,
                           node_name=self.dashboards['node_dashboard']['node_name_widget'].value,
                           site_name=self.dashboards['node_dashboard']['site_name_widget'].value,
                           cores=self.dashboards['node_dashboard']['core_slider'].value,
                           ram=self.dashboards['node_dashboard']['ram_slider'].value,
                           disk=self.dashboards['node_dashboard']['disk_slider'].value,
                           image=self.dashboards['node_dashboard']['image_widget'].value,
                           image_type=self.dashboards['node_dashboard']['image_type_widget'].value)
        else:
            print("XXX Skiping save node b/c current_node = {}".format(self.current_node))

        node_name = self.get_unique_node_name()
        self.add_node(node_name=node_name)
        self.load_node(node_name=node_name)
        self.rebuild_node_dashboard(node_name=node_name)

    def node_dashboard_delete_node(self, button, **kwargs):
        """
        Node dashboard delete node
        :param button:
        :param kwargs:
        :return:
        """
        super().delete_node(node_name=self.dashboards['node_dashboard']['node_name_widget'].value)
        self.current_node = None
        self.rebuild_node_dashboard()

    def delete_experiment(self, slice_name):
        """
        Delete a slice
        :param slice_name:
        :return:
        """
        print('Delete Slice')
        super().delete_experiment(self.current_experiment)
        self.update_slice_list()


    def update_node_dashboard(self):
        """
        Update node dashboard
        :return:
        """
        pass

    def update_link_dashboard(self):
        """
        Update link dashboard
        :return:
        """
        pass

    def update_slice_list(self, current_slice_name=None, excludes=[SliceState.Dead, SliceState.Closing]) -> List[str]:
        """
        Update Slice list
        :param current_slice_name:
        :param excludes:
        :return:
        """
        super().update_experiment_list(current_slice_name=current_slice_name, excludes=excludes)

        # Build slice name list for widget
        self.current_experiment = None
        slice_names = []
        for experiment in self.experiments:
            slice_names.append(experiment['slice_name'])
            if current_slice_name is not None and experiment['slice_name'] == current_slice_name:
                self.current_experiment = experiment

        if len(slice_names) == 0:
            slice_names = [self.DEFAULT_SLICE_SELECT_VALUE]

        # If there is no slice to set then set the first one in the list
        if current_slice_name is None and self.experiments is not None and len(self.experiments) > 0:
            current_slice_name = sorted(slice_names)[0]

        if self.experiments is not None and len(self.experiments) > 0:
            self.current_experiment = list(filter(lambda experiment: experiment['slice_name'] == current_slice_name,
                                                  self.experiments))[0]


        self.dashboards['slice_dashboard']['slice_select_widget'].options = sorted(slice_names)
        self.dashboards['slice_dashboard']['slice_select_widget'].value = current_slice_name

        if self.current_experiment is not None:
           if 'slice' in self.current_experiment.keys():
               slice = self.current_experiment['slice']
               self.dashboards['slice_dashboard']['slice_status_slice_name_value'].value = slice.slice_name
               self.dashboards['slice_dashboard']['slice_status_slice_id_value'].value = slice.slice_id
               self.dashboards['slice_dashboard']['slice_status_slice_state_value'].value = slice.slice_state
               self.dashboards['slice_dashboard']['slice_status_lease_end_value'].value = slice.lease_end
               self.dashboards['slice_dashboard']['slice_status_graph_id_value'].value = slice.graph_id
               self.dashboards['slice_dashboard']['slice_experiment_state_value'].value = self.current_experiment['editor_slice_state']
           else:
               self.dashboards['slice_dashboard']['slice_status_slice_name_value'].value = self.current_experiment['slice_name']
               self.dashboards['slice_dashboard']['slice_status_slice_state_value'].value = "new slice"
               self.dashboards['slice_dashboard']['slice_status_slice_id_value'].value = ""
               self.dashboards['slice_dashboard']['slice_status_lease_end_value'].value = ""
               self.dashboards['slice_dashboard']['slice_status_graph_id_value'].value = ""
               self.dashboards['slice_dashboard']['slice_experiment_state_value'].value = self.current_experiment['editor_slice_state']


        return slice_names

    def submit_slice(self):
        """
        Submit Slice
        :return:
        """
        self.update_slice_list()
        pass

    def open_slice_from_file(self, file_name):
        """
        Open existing slice from a file
        :param file_name:
        :return:
        """
        print("TODO open_slice_from_file")

    def save_slice_to_file(self, file_name):
        """
        Save slice to a file
        :param file_name:
        :return:
        """
        print("TODO save_slice_to_file")

    def slice_dashboard_select_slice(self, change, **kwargs):
        """
        Slice dashboard select slice
        :param change:
        :param kwargs:
        :return:
        """
        self.update_slice_list(current_slice_name=change.new)
        self.redraw_map()

    def slice_dashboard_update_slices(self, button, **kwargs):
        """
        Slice dashboard update slice
        :param button:
        :param kwargs:
        :return:
        """
        current_slice_name = None
        if self.current_experiment:
            current_slice_name = self.current_experiment['slice'].slice_name

        self.update_slice_list(current_slice_name=current_slice_name)

    def slice_dashboard_open_slice(self, button, **kwargs):
        """
        Slice dashboard open slice
        :param button:
        :param kwargs:
        :return:
        """
        self.open_slice_from_file("GET FILE NAME/PATH FROM DASHBOARD")

    def slice_dashboard_new_slice(self, button, **kwargs):
        """
        Slice dashboard new slice
        :param button:
        :param kwargs:
        :return:
        """
        new_slice_name = self.dashboards['slice_dashboard']['slice_name_widget'].value

        super().create_experiment(new_slice_name)

        self.update_slice_list(current_slice_name=new_slice_name)
        # unset the new slice name
        self.dashboards['slice_dashboard']['slice_name_widget'].value = ''

    def slice_dashboard_delete_experiment(self, button, **kwargs):
        """
        Slice dashboard delete slice
        :param button:
        :param kwargs:
        :return:
        """
        self.delete_experiment(self.dashboards['slice_dashboard']['slice_select_widget'].value)

    def slice_dashboard_submit_slice(self, button, **kwargs):
        """
        Slice dashboard submit slice
        :param button:
        :param kwargs:
        :return:
        """
        super().submit_slice(self.current_experiment)

        self.update_slice_list(current_slice_name=self.current_experiment['slice_name'])

    def node_dashboard_add_component(self, button, **kwargs):
        """
        Node dashboard add component
        :param button:
        :param kwargs:
        :return:
        """
        # Add a compopnent to the current node dashboard
        print("node_dashboard_add_component")
        component_name= self.get_unique_component_name()
        model_type=self.DEFAULT_COMPONENT_MODEL
        super().add_component(component_name, model_type)


        self.rebuild_node_dashboard()


    def handle_draw(self, info, **kwargs):
        """
        Handle draw
        :param info:
        :param kwargs:
        :return:
        """
        print("handle_draw")
        print(kwargs)



    def start(self):
        """
        Start the geo editors
        :return:
        """
        # TODO make site detail per slice and per site
        self.site_detail = False


        print('Get Available Resources')
        self.update_sites()

        print('Getting current slices')
        self.update_slice_list()

        # cyTest
        # self.cyTest()

        # help(self.dashboards['node_dashboard']['site_name_widget'])

        return self.canvas



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author: Paul Ruth (pruth@renci.org)

import os
import logging
import random

from IPython import get_ipython

from typing import List, Dict

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from fabric_cf.orchestrator.swagger_client import Slice as OrchestratorSlice, Sliver as OrchestratorSliver

from fabrictestbed.slice_manager import SliceManager, Status, SliceState
from fim.user import Node as FimNode

from fabrictestbed_extensions.fablib.abc_fablib import AbcFabLIB
from fabrictestbed_extensions.fablib.resources import Resources
from fabrictestbed_extensions.fablib.slice import Slice


class fablib(AbcFabLIB):

    log_level = logging.INFO

    def __init__(self):
        """
        Constructor. Builds SliceManager for fablib object.
        """
        super().__init__()

        self.slice_manager = None
        self.build_slice_manager()
        self.resources = None

    def build_slice_manager(self):
        """
        Not a user facing API call.

        Creates a new SliceManager object.

        :return: a new SliceManager
        :rtype: SliceManager
        """
        try:
            self.slice_manager = SliceManager(oc_host=self.orchestrator_host,
                                              cm_host=self.credmgr_host,
                                              project_id=self.project_id,
                                              scope='all')

            # Initialize the slice manager
            self.slice_manager.initialize()
        except Exception as e:
            logging.error(f"{e}")

        return self.slice_manager

    @staticmethod
    def get_image_names() -> List[str]:
        """
        Gets a list of available image names.

        :return: list of image names as strings
        :rtype: List[String]
        """
        return [ 'default_centos8_stream' ,
                 'default_centos9_stream',
                 'default_centos_7',
                 'default_centos_8',
                 'default_cirros',
                 'default_debian_10',
                 'default_fedora_35',
                 'default_freebsd_13_zfs',
                 'default_openbsd_7',
                 'default_rocky_8' ,
                 'default_ubuntu_18',
                 'default_ubuntu_20',
                 'default_ubuntu_21',
                 'default_ubuntu_22'  ]


    @staticmethod
    def get_site_names() -> List[str]:
        """
        Gets a list of all available site names.

        :return: list of site names as strings
        :rtype: List[String]
        """
        return fablib.get_resources().get_site_names()

    @staticmethod
    def list_sites() -> str:
        """
        Get a string used to print a tabular list of sites with state

        :return: tabulated string of site state
        :rtype: String
        """
        return str(fablib.get_resources())

    @staticmethod
    def show_site(site_name: str) -> str:
        """
        Get a string used to print tabular info about a site

        :param site_name: the name of a site
        :type site_name: String
        :return: tabulated string of site state
        :rtype: String
        """
        return str(fablib.get_resources().show_site(site_name))

    @staticmethod
    def get_resources() -> Resources:
        """
        Get a reference to the resourcs object. The resouces obeject
        is used to query for availale resouces and capacities.

        :return: the resouces object
        :rtype: Resources
        """
        if not fablib.fablib_object.resources:
            fablib.get_available_resources()

        return fablib.fablib_object.resources

    @staticmethod
    def get_random_site(avoid: List[str] = []) -> str:
        """
        Get a random site.

        :param avoid: list of site names to avoid chosing
        :type site_name: List[String]
        :return: one site name
        :rtype: String
        """
        return fablib.get_random_sites(count=1, avoid=avoid)[0]

    @staticmethod
    def get_random_sites(count: int = 1, avoid: List[str] = []) -> List[str]:
        """
        Get a list of random sites names. Each site will be included at most once.

        :param count: number of sites to return.
        :type count: int
        :param avoid: list of site names to avoid chosing
        :type site_name: List[String]
        :return: list of random site names.
        :rtype: List[Sting]
        """
        # Need to avoid SALT and MASS for now.
        # Real fix is to check availability
        always_avoid=['SALT', 'MASS']

        for site in always_avoid:
            if site not in avoid:
                avoid.append(site)

        sites = fablib.get_resources().get_site_list()
        for site in avoid:
            if site in sites:
                sites.remove(site)

        rtn_sites = []
        for i in range(count):
            rand_site = random.choice(sites)
            sites.remove(rand_site)
            rtn_sites.append(rand_site)
        return rtn_sites

    @staticmethod
    def init_fablib():
        """
        Not intended to be called by the user.

        Static initializer for the fablib object.
        """
        if not hasattr(fablib, 'fablib_object'):
            fablib.fablib_object = fablib()

    @staticmethod
    def get_default_slice_key() -> Dict[str, str]:
        """
        Gets the current default_slice_keys as a dictionary containg the
        public and private slice keys.

        Important! Slice key management is underdevelopment and this
        functionality will likely change going forward.

        :return: default_slice_key dictionary from superclass
        :rtype: Dict[String, String]
        """
        return fablib.fablib_object.default_slice_key

    @staticmethod
    def get_config() -> Dict[str, str]:
        """
        Gets a dictionary mapping keywords to configured FABRIC environment
        variable values values.

        :return: dictionary mapping keywords to FABRIC values
        :rtype: Dict[String, String]
        """
        return {'credmgr_host': fablib.fablib_object.credmgr_host,
                'orchestrator_host': fablib.fablib_object.orchestrator_host,
                'fabric_token': fablib.fablib_object.fabric_token,
                'bastion_username': fablib.fablib_object.bastion_username,
                'bastion_key_filename': fablib.fablib_object.bastion_key_filename,
                'bastion_public_addr': fablib.fablib_object.bastion_public_addr,
                'bastion_private_ipv4_addr': fablib.fablib_object.bastion_private_ipv4_addr,
                'slice_public_key': fablib.get_default_slice_public_key(),
                'slice_public_key_file': fablib.get_default_slice_public_key_file(),
                'slice_private_key_file': fablib.get_default_slice_private_key_file(),
                'fabric_slice_private_key_passphrase': fablib.get_default_slice_private_key_passphrase()
                }

    @staticmethod
    def get_default_slice_public_key() -> str or None:
        """
        Gets the default slice public key.

        Important! Slice key management is underdevelopment and this
        functionality will likely change going forward.

        :return: the slice public key on this fablib object
        :rtype: String
        """
        if 'slice_public_key' in fablib.fablib_object.default_slice_key.keys():
            return fablib.fablib_object.default_slice_key['slice_public_key']
        else:
            return None

    @staticmethod
    def get_default_slice_public_key_file() -> str or None:
        """
        Gets the path to the default slice public key file.

        Important! Slice key management is underdevelopment and this
        functionality will likely change going forward.

        :return: the path to the slice public key on this fablib object
        :rtype: String
        """
        if 'slice_public_key_file' in fablib.fablib_object.default_slice_key.keys():
            return fablib.fablib_object.default_slice_key['slice_public_key_file']
        else:
            return None

    @staticmethod
    def get_default_slice_private_key_file() -> str or None:
        """
        Gets the path to the default slice private key file.

        Important! Slices key management is underdevelopment and this
        functionality will likely change going forward.

        :return: the path to the slice private key on this fablib object
        :rtype: String
        """
        if 'slice_private_key_file' in fablib.fablib_object.default_slice_key.keys():
            return fablib.fablib_object.default_slice_key['slice_private_key_file']
        else:
            return None

    @staticmethod
    def get_default_slice_private_key_passphrase() -> str or None:
        """
        Gets the passphrase to the default slice private key.

        Important! Slices key management is underdevelopment and this
        functionality will likely change going forward.

        :return: the passphrase to the slice private key on this fablib object
        :rtype: String
        """
        if 'slice_private_key_passphrase' in fablib.fablib_object.default_slice_key.keys():
            return fablib.fablib_object.default_slice_key['slice_private_key_passphrase']
        else:
            return None

    @staticmethod
    def get_credmgr_host() -> str:
        """
        Gets the credential manager host site value.

        :return: the credential manager host site
        :rtype: String
        """
        return fablib.fablib_object.credmgr_host

    @staticmethod
    def get_orchestrator_host() -> str:
        """
        Gets the orchestrator host site value.

        :return: the orchestrator host site
        :rtype: String
        """
        return fablib.fablib_object.orchestrator_host

    @staticmethod
    def get_fabric_token() -> str:
        """
        Gets the FABRIC token location.

        :return: FABRIC token location
        :rtype: String
        """
        return fablib.fablib_object.fabric_token

    @staticmethod
    def get_bastion_username() -> str:
        """
        Gets the FABRIC Bastion username.

        :return: FABRIC Bastion username
        :rtype: String
        """
        return fablib.fablib_object.bastion_username

    @staticmethod
    def get_bastion_key_filename() -> str:
        """
        Gets the FABRIC Bastion key filename.

        :return: FABRIC Bastion key filename
        :rtype: String
        """
        return fablib.fablib_object.bastion_key_filename

    @staticmethod
    def get_bastion_public_addr() -> str:
        """
        Gets the FABRIC Bastion host address.

        :return: Bastion host public address
        :rtype: String
        """
        return fablib.fablib_object.bastion_public_addr

    @staticmethod
    def get_bastion_private_ipv4_addr() -> str:
        """
        Gets the FABRIC Bastion private IPv4 host address. This is the
        internally faceing IPv4 address needed to use paramiko

        :return: Bastion private IPv4 address
        :rtype: String
        """
        return fablib.fablib_object.bastion_private_ipv4_addr

    @staticmethod
    def get_bastion_private_ipv6_addr() -> str:
        """
        Gets the FABRIC Bastion private IPv6 host address. This is the
        internally faceing IPv4 address needed to use paramiko

        :return: Bastion private IPv6 address
        :rtype: String
        """
        return fablib.fablib_object.bastion_private_ipv6_addr

    @staticmethod
    def set_slice_manager(slice_manager: SliceManager) -> None:
        """
        Not intended as API call

        Sets the slice manager of this fablib object.

        :param slice_manager: the slice manager to set
        :type slice_manager: SliceManager
        """
        fablib.fablib_object.slice_manager = slice_manager

    @staticmethod
    def get_slice_manager() -> SliceManager:
        """
        Not intended as API call


        Gets the slice manager of this fablib object.

        :return: the slice manager on this fablib object
        :rtype: SliceManager
        """
        return fablib.fablib_object.slice_manager

    @staticmethod
    def new_slice(name: str) -> Slice:
        """
        Creates a new slice with the given name.

        :param name: the name to give the slice
        :type name: String
        :return: a new slice
        :rtype: Slice
        """
        # fabric = fablib()
        return Slice.new_slice(name=name)

    @staticmethod
    def get_site_advertisment(site: str) -> FimNode:
        """
        Not intended for API use.

        Given a site name, gets fim topology object for this site.

        :param site: a site name
        :type site: String
        :return: fim object for this site
        :rtype: Node
        """
        return_status, topology = fablib.get_slice_manager().resources()
        if return_status != Status.OK:
            raise Exception("Failed to get advertised_topology: {}, {}".format(return_status, topology))

        return topology.sites[site]

    @staticmethod
    def get_available_resources(update: bool = False) -> Resources:
        """
        Get the available resources.

        Optionally update the availalbe resources by querying the FABRIC
        services. Otherwise, this method returns the exisitng information.

        :param update: if True, update the available resources
        :type update: bool
        :return: Availalbe Resources object
        :rtype: Resources
        """
        from fabrictestbed_extensions.fablib.resources import Resources

        if fablib.fablib_object.resources == None:
            fablib.fablib_object.resources = Resources()

        if update:
            fablib.fablib_object.resources.update()

        return fablib.fablib_object.resources

    @staticmethod
    def get_fim_slice(excludes: List[SliceState] = [SliceState.Dead,SliceState.Closing]) -> List[OrchestratorSlice]:
        """
        Not intended for API use.

        Gets a list of fim slices from the slice manager.

        By default this method ignores Dead and Closing slices. Optional,
        parameter allows excluding a different list of slice states.  Pass
        an empty list (i.e. excludes=[]) to get a list of all slices.

        :param excludes: A list of slice states to exclude from the output list
        :type excludes: List[SliceState]
        :return: a list of slices
        :rtype: List[Slice]
        """
        return_status, slices = fablib.get_slice_manager().slices(excludes=excludes)

        return_slices = []
        if return_status == Status.OK:
            for slice in slices:
                return_slices.append(slice)
        else:
            raise Exception(f"Failed to get slice list: {slices}")
        return return_slices

    @staticmethod
    def get_slices(excludes: List[SliceState] = [SliceState.Dead,SliceState.Closing]) -> List[Slice]:
        """
        Gets a list of slices from the slice manager.

        By default this method ignores Dead and Closing slices. Optional,
        parameter allows excluding a different list of slice states.  Pass
        an empty list (i.e. excludes=[]) to get a list of all slices.

        :param excludes: A list of slice states to exclude from the output list
        :type excludes: List[SliceState]
        :return: a list of slices
        :rtype: List[Slice]
        """
        from fabrictestbed_extensions.fablib.slice import Slice
        import time

        if fablib.get_log_level() == logging.DEBUG:
            start = time.time()

        return_status, slices = fablib.get_slice_manager().slices(excludes=excludes)

        if fablib.get_log_level() == logging.DEBUG:
            end = time.time()
            logging.debug(f"Running fablib.get_slice_manager().slices(): elapsed time: {end - start} seconds")

        return_slices = []
        if return_status == Status.OK:
            for slice in slices:
                return_slices.append(Slice.get_slice(sm_slice=slice, load_config=False))
        else:
            raise Exception(f"Failed to get slices: {slices}")
        return return_slices

    @staticmethod
    def get_slice(name: str = None, slice_id: str = None) -> Slice:
        """
        Gets a slice by name or slice_id. Dead and Closing slices may have
        non-unique names and must be queried by slice_id.  Slices in all other
        states are guaranteed to have unique names and can be queried by name.

        If both a name and slicd_id are provided, the slice matching the
        slice_id will be returned.

        :param name: The name of the desired slice
        :type name: String
        :param slice_id: The ID of the desired slice
        :type slice_id: String
        :raises: Exception: if slice name or slice id are not inputted
        :return: the slice, if found
        :rtype: Slice
        """
        # Get the appropriat slices list
        if slice_id:
            # if getting by slice_id consider all slices
            slices = fablib.get_slices(excludes=[])

            for slice in slices:
                if slice_id is not None and slice.get_slice_id() == slice_id:
                    return slice
        elif name:
            # if getting by name then only consider active slices
            slices = fablib.get_slices(excludes=[SliceState.Dead, SliceState.Closing])

            for slice in slices:
                if name is not None and slice.get_name() == name:
                    return slice
        else:
            raise Exception("get_slice requires slice name (name) or slice id (slice_id)")

    @staticmethod
    def delete_slice(slice_name: str = None):
        """
        Deletes a slice by name.

        :param slice_name: the name of the slice to delete
        :type slice_name: String
        """
        slice = fablib.get_slice(slice_name)
        slice.delete()

    @staticmethod
    def delete_all(progress: bool = True):
        """
        Deletes all slices on the slice manager.

        :param progress: optional progess printing to stdout
        :type progress: Bool
        """
        slices = fablib.get_slices()

        for slice in slices:
            try:
                if progress: print(f"Deleting slice {slice.get_name()}", end='')
                slice.delete()
                if progress: print(f", Success!")
            except Exception as e:
                if progress: print(f", Failed!")

    @staticmethod
    def get_log_level():
        """
        Gets the current log level for logging
        """
        return fablib.log_level

    @staticmethod
    def set_log_level(log_level):
        """
        Sets the current log level for logging

        Options:  logging.DEBUG
                  logging.INFO
                  logging.WARNING
                  logging.ERROR
                  logging.CRITICAL

        :param log_level: new log level
        :type progress: Level
        """
        fablib.log_level = log_level

    @staticmethod
    def isJupyterNotebook():
        try:
            shell = get_ipython().__class__.__name__
            if shell == 'ZMQInteractiveShell':
                return True   # Jupyter notebook or qtconsole
            elif shell == 'TerminalInteractiveShell':
                return False  # Terminal running IPython
            else:
                return False  # Other type (?)
        except NameError:
            return False


fablib.set_log_level(logging.DEBUG)
try:
    os.makedirs("/tmp/fablib")
except:
    pass
try:
    os.makedirs("/tmp/fablib/fabric_data")
except:
    pass

logging.basicConfig(filename='/tmp/fablib/fablib.log',
                    level=fablib.get_log_level(),
                    format= '[%(asctime)s] {%(pathname)s:%(lineno)d} %(levelname)s - %(message)s',
                    datefmt='%H:%M:%S')

# init fablib object
fablib.fablib_object = fablib()



import os
from fabrictestbed.slice_manager import SliceManager, Status, SliceState
import json
from fabrictestbed.slice_editor import ExperimentTopology, Capacities, ComponentType, ComponentModelType, ServiceType, Labels
import time
import paramiko
import re

import importlib.resources as pkg_resources
from typing import List

from fabrictestbed.util.constants import Constants

from .abc_test import AbcTest

from .. import images

class NetworkBencharks(AbcTest):

    def __init__(self):
        """
        Constructor
        :return:
        """
        super().__init__()


    def latency_test(self, ssh_client_n1, ssh_client_n2, dataplane_ip_n1, dataplane_ip_n2, verbose=False, info=None):
        if verbose: print("Testing Latency: {}".format(info),end='')

        #rtt min/avg/max/mdev = 0.063/0.119/0.189/0.053 ms
        #output = "Information about latency with ping: \n"

        #warm up
        stdin, stdout, stderr = ssh_client_n1.exec_command('ping -c 3 ' + dataplane_ip_n2 + ' | grep rtt')

        #Run test
        output = {}

        stdin, stdout, stderr = ssh_client_n1.exec_command('ping -c 10 ' + dataplane_ip_n2 + ' | grep rtt')
        raw_output = stdout.read().decode("utf-8")
        raw_data = raw_output.split(" ")[3]
        data_array = raw_data.split("/")
        output['rtt'] = {   'min': data_array[0],
                                'avg': data_array[1],
                                'max': data_array[2],
                                'mdev': data_array[3],
                            }
        if verbose: print(", avg rtt: {}".format(output['rtt']['avg']),end='')

        stdin, stdout, stderr = ssh_client_n2.exec_command('ping -c 10 ' + dataplane_ip_n1 + ' | grep rtt')
        #output += "\n" + stdout.read().decode("utf-8")
        raw_output = stdout.read().decode("utf-8")
        raw_data = raw_output.split(" ")[3]
        data_array = raw_data.split("/")
        output['rtt_rev'] = {   'min': data_array[0],
                                'avg': data_array[1],
                                'max': data_array[2],
                                'mdev': data_array[3],
                            }
        if verbose: print(", avg rtt_rev: {}".format(output['rtt_rev']['avg']))

        #return {'latency_test': output}
        return output

    def mtu_test(self, ssh_client_n1, ssh_client_n2, dataplane_ip_n1, dataplane_ip_n2, verbose=False, info=None):
        if verbose: print("Testing MTU: {}".format(info),end='')

        #Run test
        output = {}

        ping_packets_count = 3
        #ping_packet_sizes = [9000, 8950, 8000, 1500, 1450, 1400, 1000, 500, 100, 50]
        max_success = -1
        min_fail = 10000

        #Test min ping
        current_size = 0

        stdin, stdout, stderr = ssh_client_n1.exec_command('ping -M do -s ' + str(current_size) + ' -c ' + str(ping_packets_count) + ' ' + dataplane_ip_n2 + " | grep transmitted")
        ping_string = stdout.read().decode("utf-8")

        data_array = ping_string.split(" ")
        recieved_packets = 0
        if len(data_array) > 3:
            recieved_packets = data_array[3]

        if(int(recieved_packets) == ping_packets_count):
            max_success = current_size
            min_fail = 10000
        else:
            min_fail = 0
            max_success = -1


        while max_success < min_fail - 1:

            current_size = int((min_fail+max_success)/2)
            #print("min_fail: {}, max_success: {}, current_size: {}".format(str(min_fail),str(max_success), str(current_size)))

            stdin, stdout, stderr = ssh_client_n1.exec_command('ping -M do -s ' + str(current_size) + ' -c ' + str(ping_packets_count) + ' ' + dataplane_ip_n2 + " | grep transmitted")
            ping_string = stdout.read().decode("utf-8")

            #print("current_size: {}, output: {}".format(str(current_size),ping_string))

            data_array = ping_string.split(" ")
            recieved_packets = 0
            if len(data_array) > 3:
                recieved_packets = data_array[3]

            #print("recieved_packets: {}, ping_packets_count: {}".format(str(recieved_packets),str(ping_packets_count)))
            if(int(recieved_packets) == int(ping_packets_count)):
                max_success = current_size
            else:
                min_fail = current_size

        if max_success > 0:
            output['mtu'] = str(max_success+28)
        else:
            output['mtu'] = 0

        if verbose: print(", mtu: {}".format(output['mtu']),end='')


        #Test reverse
        max_success = -1
        min_fail = 10000

        #Test min ping
        current_size = 0

        stdin, stdout, stderr = ssh_client_n2.exec_command('ping -M do -s ' + str(current_size) + ' -c ' + str(ping_packets_count) + ' ' + dataplane_ip_n1 + " | grep transmitted")
        ping_string = stdout.read().decode("utf-8")

        data_array = ping_string.split(" ")
        recieved_packets = 0
        if len(data_array) > 3:
            recieved_packets = data_array[3]

        if(int(recieved_packets) == ping_packets_count):
            max_success = current_size
            min_fail = 10000
        else:
            min_fail = 0
            max_success = -1


        while max_success < min_fail - 1:

            current_size = int((min_fail+max_success)/2)
            #print("min_fail: {}, max_success: {}, current_size: {}".format(str(min_fail),str(max_success), str(current_size)))

            stdin, stdout, stderr = ssh_client_n2.exec_command('ping -M do -s ' + str(current_size) + ' -c ' + str(ping_packets_count) + ' ' + dataplane_ip_n1 + " | grep transmitted")
            ping_string = stdout.read().decode("utf-8")

            #print("current_size: {}, output: {}".format(str(current_size),ping_string))

            data_array = ping_string.split(" ")
            recieved_packets = 0
            if len(data_array) > 3:
                recieved_packets = data_array[3]

            #print("recieved_packets: {}, ping_packets_count: {}".format(str(recieved_packets),str(ping_packets_count)))
            if(int(recieved_packets) == int(ping_packets_count)):
                max_success = current_size
            else:
                min_fail = current_size

        if max_success > 0:
            output['mtu_rev'] = str(max_success+28)
        else:
            output['mtu_rev'] = 0

        if verbose: print(", mtu_rev: {}  ".format(output['mtu_rev']))


        #return {'mtu_test': output}
        return output


    def bandwidth_test(self, ssh_client_n1, ssh_client_n2, dataplane_ip_n1, dataplane_ip_n2, verbose=False, info=None):
        if verbose: print("Testing Bandwidth: {}".format(info),end='')

        output = {}
        #stdin, stdout, stderr = ssh_client_n1.exec_command('echo "net.core.rmem_max = 2147483647\nnet.core.wmem_max = 2147483647\nnet.ipv4.tcp_rmem = 4096 87380 2147483647\nnet.ipv4.tcp_wmem = 4096 65536 2147483647\nnet.ipv4.tcp_congestion_control=htcp\nnet.ipv4.tcp_mtu_probing=1\nnet.core.default_qdisc = fq\n" | sudo tee -a /etc/sysctl.conf && sudo sysctl -p')
        #stdin, stdout, stderr = ssh_client_n2.exec_command('echo "net.core.rmem_max = 2147483647\nnet.core.wmem_max = 2147483647\nnet.ipv4.tcp_rmem = 4096 87380 2147483647\nnet.ipv4.tcp_wmem = 4096 65536 2147483647\nnet.ipv4.tcp_congestion_control=htcp\nnet.ipv4.tcp_mtu_probing=1\nnet.core.default_qdisc = fq\n" | sudo tee -a /etc/sysctl.conf && sudo sysctl -p')

        stdin, stdout, stderr = ssh_client_n1.exec_command('echo "net.core.rmem_max = 2147483647\nnet.core.wmem_max = 2147483647\nnet.ipv4.tcp_rmem = 4096 87380 2147483647\nnet.ipv4.tcp_wmem = 4096 65536 2147483647\nnet.ipv4.tcp_congestion_control=htcp\nnet.ipv4.tcp_mtu_probing=1\nnet.core.default_qdisc = fq\n" | sudo tee -a /etc/sysctl.conf && sudo sysctl -p')
        stdin, stdout, stderr = ssh_client_n2.exec_command('echo "net.core.rmem_max = 2147483647\nnet.core.wmem_max = 2147483647\nnet.ipv4.tcp_rmem = 4096 87380 2147483647\nnet.ipv4.tcp_wmem = 4096 65536 2147483647\nnet.ipv4.tcp_congestion_control=htcp\nnet.ipv4.tcp_mtu_probing=1\nnet.core.default_qdisc = fq\n" | sudo tee -a /etc/sysctl.conf && sudo sysctl -p')


        stdin, stdout, stderr = ssh_client_n1.exec_command('iperf3 -s > /dev/null 2>&1 &')
        stdin, stdout, stderr = ssh_client_n2.exec_command('iperf3 -s > /dev/null 2>&1 &')

        stdin, stdout, stderr = ssh_client_n2.exec_command('iperf3 -J -c ' + dataplane_ip_n1 + '-t 60 -P 1 -w 512M')
        try:
            results = json.loads(str(stdout.read(),'utf-8'))
            output['forward'] = results
        except Exception as e:
            print("error {}".format(e))
            print("iperf raw results: {}".format(results))

        if verbose:
            bps = output['forward']['end']['sum_received']['bits_per_second']
            gbps = float(bps)/1000000000
            print(", forward: {:.3f} gbps".format(gbps),end='')



        stdin, stdout, stderr = ssh_client_n1.exec_command('iperf3 -J -c ' + dataplane_ip_n2 + '-t 60 -P 1 -w 512M')
        try:
            results = json.loads(str(stdout.read(),'utf-8'))
            output['reverse'] = results
        except Exception as e:
            print("error {}".format(e))
            print("iperf raw results: {}".format(results))

        if verbose:
            bps = output['reverse']['end']['sum_received']['bits_per_second']
            gbps = float(bps)/1000000000
            print(", reverse: {:.3f} gbps".format(gbps))

        #return {'bandwidth_test': output}
        return output



    def network_card_information(self, ssh_client_n1, ssh_client_n2, dataplane_ip_n1, dataplane_ip_n2, verbose):
        output = "n1\n\n"

        output += "lspci -xxxvvv\n\n"
        stdin, stdout, stderr = ssh_client_n1.exec_command('sudo lspci -xxxvvv | grep PN')
        output += stdout.read().decode("utf-8")
        stdin, stdout, stderr = ssh_client_n1.exec_command('sudo lspci -xxxvvv | grep V2')
        output += stdout.read().decode("utf-8")
        stdin, stdout, stderr = ssh_client_n1.exec_command('sudo lspci -xxxvvv | grep SN')
        output += stdout.read().decode("utf-8")
        stdin, stdout, stderr = ssh_client_n1.exec_command('sudo lspci -xxxvvv | grep PN')
        output += stdout.read().decode("utf-8")
        stdin, stdout, stderr = ssh_client_n1.exec_command('sudo lspci -xxxvvv | grep V3')
        output += stdout.read().decode("utf-8")
        stdin, stdout, stderr = ssh_client_n1.exec_command('sudo lspci -xxxvvv | grep VA')
        output += stdout.read().decode("utf-8")
        stdin, stdout, stderr = ssh_client_n1.exec_command('sudo lspci -xxxvvv | grep V0')
        output += stdout.read().decode("utf-8")

        output += "\n\n\nip a\n\n"
        stdin, stdout, stderr = ssh_client_n1.exec_command('ip a')
        output += stdout.read().decode("utf-8")

        output += "\n\n\n\nn2\n\n"

        output += "lspci -xxxvvv\n\n"
        stdin, stdout, stderr = ssh_client_n2.exec_command('sudo lspci -xxxvvv | grep PN')
        output += stdout.read().decode("utf-8")
        stdin, stdout, stderr = ssh_client_n2.exec_command('sudo lspci -xxxvvv | grep V2')
        output += stdout.read().decode("utf-8")
        stdin, stdout, stderr = ssh_client_n2.exec_command('sudo lspci -xxxvvv | grep SN')
        output += stdout.read().decode("utf-8")
        stdin, stdout, stderr = ssh_client_n2.exec_command('sudo lspci -xxxvvv | grep PN')
        output += stdout.read().decode("utf-8")
        stdin, stdout, stderr = ssh_client_n2.exec_command('sudo lspci -xxxvvv | grep V3')
        output += stdout.read().decode("utf-8")
        stdin, stdout, stderr = ssh_client_n2.exec_command('sudo lspci -xxxvvv | grep VA')
        output += stdout.read().decode("utf-8")
        stdin, stdout, stderr = ssh_client_n2.exec_command('sudo lspci -xxxvvv | grep V0')
        output += stdout.read().decode("utf-8")

        output += "\n\n\nip a\n\n"
        stdin, stdout, stderr = ssh_client_n2.exec_command('ip a')
        output += stdout.read().decode("utf-8")

        return {"NetworkCardInformation" : output}

    def processor_information(self, ssh_client_n1, ssh_client_n2, dataplane_ip_n1, dataplane_ip_n2, verbose):
        output = "n1\n\n"

        output += 'sudo dmidecode | grep -w ID | sed "s/^.ID\: //g"'
        stdin, stdout, stderr = ssh_client_n1.exec_command('sudo dmidecode | grep -w ID | sed "s/^.ID\: //g"')
        output += stdout.read().decode("utf-8")

        output += '\n\n\ndmesg | grep -i dmi: | cut -d ":" -f 2-\n\n'
        stdin, stdout, stderr = ssh_client_n1.exec_command('dmesg | grep -i dmi: | cut -d ":" -f 2-')
        output += stdout.read().decode("utf-8")

        output += "\n\n\nsudo dmidecode -s system-serial-number\n\n"
        stdin, stdout, stderr = ssh_client_n1.exec_command('sudo dmidecode -s system-serial-number')
        output += stdout.read().decode("utf-8")

        output += "\n\n\n\nn2\n\n"

        output += 'sudo dmidecode | grep -w ID | sed "s/^.ID\: //g"'
        stdin, stdout, stderr = ssh_client_n2.exec_command('sudo dmidecode | grep -w ID | sed "s/^.ID\: //g"')
        output += stdout.read().decode("utf-8")

        output += '\n\n\ndmesg | grep -i dmi: | cut -d ":" -f 2-\n\n'
        stdin, stdout, stderr = ssh_client_n2.exec_command('dmesg | grep -i dmi: | cut -d ":" -f 2-')
        output += stdout.read().decode("utf-8")

        output += "\n\n\nsudo dmidecode -s system-serial-number\n\n"
        stdin, stdout, stderr = ssh_client_n2.exec_command('sudo dmidecode -s system-serial-number')
        output += stdout.read().decode("utf-8")

        return {"ProcessorInformation" : output}



    def create_ptp_test_slice(self, test=None, verbose = True):

        slice_name = test['test_name']
        site1 = test['src']['site']
        site2 = test['dst']['site']


        #Create Topo
        t = ExperimentTopology()

        #Node1
        cap = Capacities(core=test['src']['core'], ram=test['src']['ram'], disk=test['src']['disk'])
        n1 = t.add_node(name='node1', site=site1)
        n1.set_properties(capacities=cap, image_type='qcow2', image_ref='default_ubuntu_20')
        if test['src']['nic'] == 'SmartNIC_ConnectX_6':
            n1.add_component(model_type=ComponentModelType.SmartNIC_ConnectX_6, name='n1-nic1')
        elif test['src']['nic'] == 'SmartNIC_ConnectX_5':
            n1.add_component(model_type=ComponentModelType.SmartNIC_ConnectX_5, name='n1-nic1')
        elif test['src']['nic'] == 'SharedNIC_ConnectX_6':
            n1.add_component(model_type=ComponentModelType.SharedNIC_ConnectX_6, name='n1-nic1')


        #Node2
        cap = Capacities(core=test['dst']['core'], ram=test['dst']['ram'], disk=test['dst']['disk'])
        n2 = t.add_node(name='node2', site=site2)
        n2.set_properties(capacities=cap, image_type='qcow2', image_ref='default_ubuntu_20')
        if test['dst']['nic'] == 'SmartNIC_ConnectX_6':
            n2.add_component(model_type=ComponentModelType.SmartNIC_ConnectX_6, name='n2-nic1')
        elif test['dst']['nic'] == 'SmartNIC_ConnectX_5':
            n2.add_component(model_type=ComponentModelType.SmartNIC_ConnectX_5, name='n2-nic1')
        elif test['dst']['nic'] == 'SharedNIC_ConnectX_6':
            n2.add_component(model_type=ComponentModelType.SharedNIC_ConnectX_6, name='n2-nic1')


        # Network
        t.add_network_service(name='ptp1', nstype=ServiceType.L2PTP, interfaces=[n1.interface_list[0], n2.interface_list[0]])
        if_labels = n1.interface_list[0].get_property(pname="labels")
        if_labels.vlan = "200"
        n1.interface_list[0].set_properties(labels=if_labels)
        if_labels = n2.interface_list[0].get_property(pname="labels")
        if_labels.vlan = "200"
        n2.interface_list[0].set_properties(labels=if_labels)

        #Submit
        slice_graph = t.serialize()

        status, reservations = self.slice_manager .create(slice_name=slice_name, slice_graph=slice_graph, ssh_key=self.node_ssh_key)
        if(status != Status.OK):
            print(status)
            print(reservations)
            raise Exception("Slice creation failed.")
        slice_id=reservations[0].slice_id
        print("slice_id: {}".format(slice_id))


        time.sleep(10)
        return_status, slices = self.slice_manager .slices(excludes=[SliceState.Dead,SliceState.Closing])
        if(return_status != Status.OK):
            print(return_status)
            print(slices)
            raise Exception("Slice get failed.")
        slice = list(filter(lambda x : x.slice_name == slice_name, slices))[0]

        return slice

    def configure_test_node(self, node, dataplane_ip=None, vlan=None, verbose = True):

        node_ip = node.management_ip

        if(verbose):
            print('Server: {}'.format(node.name))
            print("   Cores             : {}".format(node.get_property(pname='capacity_allocations').core))
            print("   RAM               : {}".format(node.get_property(pname='capacity_allocations').ram))
            print("   Disk              : {}".format(node.get_property(pname='capacity_allocations').disk))
            print("   Image             : {}".format(node.image_ref))
            print("   Host              : {}".format(node.get_property(pname='label_allocations').instance_parent))
            print("   Site              : {}".format(node.site))
            print("   Management IP     : {}".format(node.management_ip))
            print("   Components        :")
            for component_name, component in node.components.items():
                print("      Name             : {}".format(component.name))
                print("      Model            : {}".format(component.model))
                print("      Type             : {}".format(component.type))

        #Configure
        ssh_client = self.open_ssh_client('ubuntu', node)

        #Configure node1
        stdin, stdout, stderr = ssh_client.exec_command('sudo apt-get update && sudo apt-get install -y iperf iperf3')
        stdin, stdout, stderr = ssh_client.exec_command('ip -j a')
        dataplane_iface = self.get_dataplane_interface(str(stdout.read(),'utf-8').replace('\\n','\n'))
        print("dataplane_iface: {}".format(dataplane_iface))

        if vlan:
            #dataplane_ip_n1 = "192.168.10.51"
            stdin, stdout, stderr = ssh_client.exec_command('sudo ip link add link ' + dataplane_iface + ' name ens7.200 type vlan id 200')
            stdin, stdout, stderr = ssh_client.exec_command('sudo ip link set dev ' + dataplane_iface + ' up mtu 9000')
            stdin, stdout, stderr = ssh_client.exec_command('sudo ip link set dev ens7.200 up mtu 9000')
            stdin, stdout, stderr = ssh_client.exec_command('sudo ip addr add ' + dataplane_ip + '/24 dev ens7.200')
        else:
            stdin, stdout, stderr = ssh_client.exec_command('sudo ip link set dev ens7 up mtu 9000')
            stdin, stdout, stderr = ssh_client.exec_command('sudo ip addr add ' + dataplane_ip + '/24 dev ens7')

        self.close_ssh_client(ssh_client)



    def run_tests(self, test_name, node1, node2, dataplane_ip_n1, dataplane_ip_n2, test_list, verbose = True):

        #Configure
        ssh_client_n1 = self.open_ssh_client('ubuntu', node1)
        ssh_client_n2 = self.open_ssh_client('ubuntu', node2)

        output = {}
        for test in test_list:
            #print("{}".format(str(test)))
            #print("running test")
            try:
                output[test.__name__] = test(ssh_client_n1, ssh_client_n2, dataplane_ip_n1, dataplane_ip_n2, verbose=verbose, info="{}-{}".format(node1.name,node2.name))
            except Exception as e:
                print("Exception running test: {}".format(str(e)))


        #self.slice_manager .delete(slice_object=slice_object)

        self.close_ssh_client(ssh_client_n1)
        self.close_ssh_client(ssh_client_n2)

        #if(verbose):
            #for k in output:
                #print(k)
                #print(output[k])

        return output

    def run_tests_all_tests(self, test_name, site1, site2, verbose=False):
        credmgr_host = os.environ[Constants.FABRIC_CREDMGR_HOST]
        orchestrator_host = os.environ[Constants.FABRIC_ORCHESTRATOR_HOST]
        project_id = os.environ[Constants.FABRIC_PROJECT_ID]
        self.slice_manager = SliceManager(oc_host=orchestrator_host, cm_host=credmgr_host, project_id=project_id,
                                          scope='all')
        self.slice_manager .initialize()

        slice = self.create_ptp_test_slice(test_name, site1, site2, verbose = True)

        slice = self.wait_for_slice(slice, progress=verbose, timeout=600)

        time.sleep(120)
        return_status, topology = self.slice_manager .get_slice_topology(slice_object=slice)
        if return_status != Status.OK:
            raise Exception("run_ssh_test failed to get topology. slice; {}, error {}".format(str(slice),str(topology)))

        node_num = 100
        nodes = {}
        for node_name, node in topology.nodes.items():
            dataplane_ip = '192.168.1.'+str(node_num)
            node_num = node_num + 1
            nodes[node_name] = { 'dataplane_ip': dataplane_ip, 'node': node}

            self.configure_test_node(node, dataplane_ip,vlan=200)


        #self.run_tests(site1, site2, [self.latency_test])
        #self.run_tests(test_name, site1, site2, [self.latency_test, self.mtu_test, self.bandwidth_test, self.network_card_information, self.processor_information])
        n1 = nodes['node1']
        n2 = nodes['node2']
        self.run_tests(test_name, n1['node'], n2['node'], n1['dataplane_ip'], n2['dataplane_ip'], [self.latency_test, self.mtu_test, self.bandwidth_test])

    def test_all_ptp(self, tests, verbose=False, create_slices=True):
        credmgr_host = os.environ[Constants.FABRIC_CREDMGR_HOST]
        orchestrator_host = os.environ[Constants.FABRIC_ORCHESTRATOR_HOST]
        project_id = os.environ[Constants.FABRIC_PROJECT_ID]

        slice_manager = SliceManager(oc_host=orchestrator_host, cm_host=credmgr_host, project_id=project_id,
                                     scope='all')

        # Initialize the slice manager
        slice_manager.initialize()

        #_status,resources = slice_manager.resources()

        #Get link pairs
        #site_pairs_list = []
        #for key in resources.links:
        #    site_pairs_list.append(resources.links[key].interface_list[0].name.split("_"))

        #tests = []
        #for test_pair in site_pairs:
        #    tests.append({ 'src': test_pair[0], 'dst': test_pair[1]})

        #Create Slices
        for test in tests:
            if create_slices:
                test['slice'] = self.create_ptp_test_slice(test=test, verbose = True)
            else:
                test['slice'] = self.get_slice(slice_name=test['test_name'],slice_manager=self.slice_manager)

        #wait for slices
        for test in tests:
            test['slice'] = self.wait_for_slice(test['slice'], progress=verbose, timeout=600)

        if create_slices:
            time.sleep(120)

        #Run tests
        all_results = {}
        for test in tests:
            #print("Process test: {}".format(str(test)))
            slice = test['slice']
            test_name = test['test_name']
            return_status, topology = self.slice_manager .get_slice_topology(slice_object=slice)
            if return_status != Status.OK:
                raise Exception("run_ssh_test failed to get topology. slice; {}, error {}".format(str(slice),str(topology)))

            node_num = 100
            nodes = {}
            for node_name, node in topology.nodes.items():
                dataplane_ip = '192.168.1.'+str(node_num)
                node_num = node_num + 1
                nodes[node_name] = { 'dataplane_ip': dataplane_ip, 'node': node}
                print("Config node: {}".format(str(node_name)))
                self.configure_test_node(node, dataplane_ip, vlan=200)


            #self.run_tests(site1, site2, [self.latency_test])
            #self.run_tests(test_name, site1, site2, [self.latency_test, self.mtu_test, self.bandwidth_test, self.network_card_information, self.processor_information])
            n1 = nodes['node1']
            n2 = nodes['node2']
            #print("Run test: {}".format(str(test['test_name'])))
            results = self.run_tests(test_name, n1['node'], n2['node'], n1['dataplane_ip'], n2['dataplane_ip'], [self.latency_test, self.mtu_test, self.bandwidth_test])
            all_results[test_name]=results

        return all_results


    def print_summary(self, all_results):

        print("{:<60} | {:>9} | {:>9} | {:>11} | {:>11} | {:>10} | {:>10}".format(" ","rtt","rtt_rev","mtu","mtu_rev","bw","bw_rev"))
        print("------------------------------------------------------------------------------------------------------------------------------------------")
        for test_name, results in all_results.items():
            print("{:<60}".format(test_name), end='')

            rtt=''
            rtt_rev=''
            if 'latency_test' in results.keys():
                rtt=results['latency_test']['rtt']['avg']
                rtt_rev=results['latency_test']['rtt_rev']['avg']
            print(" | {:>6} ms".format(str(rtt)), end='')
            print(" | {:>6} ms".format(str(rtt_rev)), end='')

            mtu = ''
            mtu_rev = ''
            if 'mtu_test' in results.keys():
                mtu=results['mtu_test']['mtu']
                mtu_rev=results['mtu_test']['mtu_rev']
            print(" | {:>5} bytes".format(str(mtu)), end='')
            print(" | {:>5} bytes".format(str(mtu_rev)), end='')

            gbps_forward = 0.0
            gbps_reverse = 0.0
            if 'bandwidth_test' in results.keys():
                bps_reverse=results['bandwidth_test']['reverse']['end']['sum_received']['bits_per_second']
                bps_forward=results['bandwidth_test']['forward']['end']['sum_received']['bits_per_second']
                gbps_reverse = float(bps_reverse)/1000000000
                gbps_forward = float(bps_forward)/1000000000
            print(" | {:.3f} gbps".format(gbps_forward), end='')
            print(" | {:.3f} gbps".format(gbps_reverse), end='')
            print('')






    def get_dataplane_interface(self, stdout, ip=None):
        try:
            interfaces = json.loads(str(stdout))
            for iface in  interfaces:
                if iface['link_type'] != 'loopback' and iface['operstate'] != 'UP':
                    addr_info = iface['addr_info']
                    for addr in addr_info:
                        if iface['local'] == ip:
                            return iface['ifname']
                    #print ('{}'.format(iface['ifname']))
                    break

        except Exception as e:
            print("stdout: {}".format(stdout))
            raise e
        return iface['ifname']

    def create_s2s_test_slice(self, test_name, test=None, verbose = True):
        slice_name = test_name
        site = test['server']['site']

        #Create Topo
        t = ExperimentTopology()

        interface_list = []

        #server
        server = test['server']
        cap = Capacities(core=server['core'], ram=server['ram'], disk=server['disk'])
        node = t.add_node(name=server['node_name'], site=site)

        print("Create node: {}, core: {}, ram: {}, disk: {}".format(server['node_name'],server['core'],server['ram'],server['disk']))
        node.set_properties(capacities=cap, image_type='qcow2', image_ref='default_ubuntu_20')
        if server['nic'] == 'SmartNIC_ConnectX_6':
            node.add_component(model_type=ComponentModelType.SmartNIC_ConnectX_6, name=server['node_name']+'-nic1')
        elif server['nic'] == 'SmartNIC_ConnectX_5':
            node.add_component(model_type=ComponentModelType.SmartNIC_ConnectX_5, name=server['node_name']+'-nic1')
        elif server['nic'] == 'SharedNIC_ConnectX_6':
            node.add_component(model_type=ComponentModelType.SharedNIC_ConnectX_6, name=server['node_name']+'-nic1')

        interface_list.append(node.interface_list[0])


        for client in test['clients']:
            site = client['site']
            #Node2
            cap = Capacities(core=client['core'], ram=client['ram'], disk=client['disk'])
            node = t.add_node(name=client['node_name'], site=site)

            print("Create node: {}, core: {}, ram: {}, disk: {}".format(client['node_name'],client['core'],client['ram'],client['disk']))
            node.set_properties(capacities=cap, image_type='qcow2', image_ref='default_ubuntu_20')
            if client['nic'] == 'SmartNIC_ConnectX_6':
                node.add_component(model_type=ComponentModelType.SmartNIC_ConnectX_6, name=client['node_name']+'-nic1')
            elif client['nic'] == 'SmartNIC_ConnectX_5':
                node.add_component(model_type=ComponentModelType.SmartNIC_ConnectX_5, name=client['node_name']+'-nic1')
            elif client['nic'] == 'SharedNIC_ConnectX_6':
                node.add_component(model_type=ComponentModelType.SharedNIC_ConnectX_6, name=client['node_name']+'-nic1')
            else:
                print("Error setting nic type: node: {}, nic: {}".format(client['node_name'],client['nic']))

            interface_list.append(node.interface_list[0])

        # Network
        t.add_network_service(name='l2br1', nstype=ServiceType.L2STS, interfaces=interface_list)

        #Submit
        slice_graph = t.serialize()

        status, reservations = self.slice_manager.create(slice_name=slice_name, slice_graph=slice_graph, ssh_key=self.node_ssh_key)
        if(status != Status.OK):
            print(status)
            print(reservations)
            raise Exception("Slice creation failed.")
        slice_id=reservations[0].slice_id
        print("slice_id: {}".format(slice_id))


        time.sleep(10)
        return_status, slices = self.slice_manager.slices(excludes=[SliceState.Dead,SliceState.Closing])
        if(return_status != Status.OK):
            print(return_status)
            print(slices)
            raise Exception("Slice get failed.")
        slice = list(filter(lambda x : x.slice_name == slice_name, slices))[0]

        return slice

    def create_l2bridge_test_slice(self, test_name, test=None, verbose = True):
        slice_name = test_name
        site = test['site']

        #Create Topo
        t = ExperimentTopology()

        interface_list = []

        #server
        server = test['server']
        cap = Capacities(core=server['core'], ram=server['ram'], disk=server['disk'])
        node = t.add_node(name=server['node_name'], site=site)

        #temp for testign
        #labels = Labels()
        #labels.instance_parent = 'star-w5.fabric-testbed.net'


        print("Create node: {}, core: {}, ram: {}, disk: {}".format(server['node_name'],server['core'],server['ram'],server['disk']))
        #node.set_properties(capacities=cap, image_type='qcow2', image_ref='default_ubuntu_20',labels=labels)
        node.set_properties(capacities=cap, image_type='qcow2', image_ref='default_ubuntu_20')
        if server['nic'] == 'SmartNIC_ConnectX_6':
            node.add_component(model_type=ComponentModelType.SmartNIC_ConnectX_6, name=server['node_name']+'-nic1')
        elif server['nic'] == 'SmartNIC_ConnectX_5':
            node.add_component(model_type=ComponentModelType.SmartNIC_ConnectX_5, name=server['node_name']+'-nic1')
        elif server['nic'] == 'SharedNIC_ConnectX_6':
            node.add_component(model_type=ComponentModelType.SharedNIC_ConnectX_6, name=server['node_name']+'-nic1')

        interface_list.append(node.interface_list[0])


        for client in test['clients']:

            #Node2
            cap = Capacities(core=client['core'], ram=client['ram'], disk=client['disk'])
            node = t.add_node(name=client['node_name'], site=site)

            #labels = Labels()
            #labels.instance_parent = 'star-w4.fabric-testbed.net'


            print("Create node: {}, core: {}, ram: {}, disk: {}".format(client['node_name'],client['core'],client['ram'],client['disk']))
            #node.set_properties(capacities=cap, image_type='qcow2', image_ref='default_ubuntu_20',labels=labels)
            node.set_properties(capacities=cap, image_type='qcow2', image_ref='default_ubuntu_20')
            if client['nic'] == 'SmartNIC_ConnectX_6':
                node.add_component(model_type=ComponentModelType.SmartNIC_ConnectX_6, name=client['node_name']+'-nic1')
            elif client['nic'] == 'SmartNIC_ConnectX_5':
                node.add_component(model_type=ComponentModelType.SmartNIC_ConnectX_5, name=client['node_name']+'-nic1')
            elif client['nic'] == 'SharedNIC_ConnectX_6':
                node.add_component(model_type=ComponentModelType.SharedNIC_ConnectX_6, name=client['node_name']+'-nic1')

            interface_list.append(node.interface_list[0])

        # Network
        t.add_network_service(name='l2br1', nstype=ServiceType.L2Bridge, interfaces=interface_list)

        #Submit
        slice_graph = t.serialize()

        status, reservations = self.slice_manager.create(slice_name=slice_name, slice_graph=slice_graph, ssh_key=self.node_ssh_key)
        if(status != Status.OK):
            print(status)
            print(reservations)
            raise Exception("Slice creation failed.")
        slice_id=reservations[0].slice_id
        print("slice_id: {}".format(slice_id))


        time.sleep(10)
        return_status, slices = self.slice_manager.slices(excludes=[SliceState.Dead,SliceState.Closing])
        if(return_status != Status.OK):
            print(return_status)
            print(slices)
            raise Exception("Slice get failed.")
        slice = list(filter(lambda x : x.slice_name == slice_name, slices))[0]

        return slice

    def test_s2s(self, test_name, test=None, verbose=False, create_slice=True):
        credmgr_host = os.environ[Constants.FABRIC_CREDMGR_HOST]
        orchestrator_host = os.environ[Constants.FABRIC_ORCHESTRATOR_HOST]
        project_id = os.environ[Constants.FABRIC_PROJECT_ID]
        self.slice_manager = SliceManager(oc_host=orchestrator_host, cm_host=credmgr_host, project_id=project_id,
                                          scope='all')
        self.slice_manager.initialize()

        if create_slice:
            slice = self.create_s2s_test_slice(test_name, test=test, verbose=verbose)
        else:
            slice = self.get_slice(slice_name=test_name,slice_manager=self.slice_manager)

        slice = self.wait_for_slice(slice, progress=verbose, timeout=600)

        if create_slice:
            time.sleep(120)

        return_status, topology = self.slice_manager.get_slice_topology(slice_object=slice)
        if return_status != Status.OK:
            raise Exception("run_ssh_test failed to get topology. slice; {}, error {}".format(str(slice),str(topology)))

        #print("topology: {}".format(str(topology)))
        server_node = topology.nodes[test['server']['node_name']]
        server_name = server_node.name
        server_dataplane_ip='192.168.1.1'
        print("Config server: {}".format(str(server_name)))
        self.configure_test_node(server_node, server_dataplane_ip)

        all_results = {}

        node_num = 100
        nodes = []
        for node_name, node in topology.nodes.items():
            if node_name == server_name:
                continue

            dataplane_ip = '192.168.1.'+str(node_num)
            node_num = node_num + 1
            nodes.append({ 'dataplane_ip': dataplane_ip, 'node': node})
            print("Config node: {}".format(str(node_name)))
            self.configure_test_node(node, dataplane_ip)


        for node in nodes:

            result_name = "{}_{}_{}".format(test_name, server_name, node['node'].name)
            #self.run_tests(site1, site2, [self.latency_test])
            #self.run_tests(test_name, site1, site2, [self.latency_test, self.mtu_test, self.bandwidth_test, self.network_card_information, self.processor_information])
            #print("Run test: {}".format(str(server_node)))
            #print("Run test: {}".format(str(node['node'])))
            #print("Run test: {}".format(str(server_dataplane_ip)))
            #print("Run test: {}".format(str(node['dataplane_ip'])))
            #print("Run test: {} {} {} {}".format(str(server_node), str(node['node']), str(server_dataplane_ip), str(node['dataplane_ip'])))
            all_results[result_name] = self.run_tests(test_name, server_node, node['node'], server_dataplane_ip, node['dataplane_ip'], [self.latency_test, self.mtu_test, self.bandwidth_test])


        return all_results


    def test_l2bridge(self, test_name, test=None, verbose=False, create_slice=True):
        credmgr_host = os.environ[Constants.FABRIC_CREDMGR_HOST]
        orchestrator_host = os.environ[Constants.FABRIC_ORCHESTRATOR_HOST]
        project_id = os.environ[Constants.FABRIC_PROJECT_ID]
        self.slice_manager = SliceManager(oc_host=orchestrator_host, cm_host=credmgr_host, project_id=project_id,
                                          scope='all')
        self.slice_manager.initialize()

        site = test['site']

        if create_slice:
            slice = self.create_l2bridge_test_slice(test_name, test=test, verbose=verbose)
        else:
            slice = self.get_slice(slice_name=test_name,slice_manager=self.slice_manager)

        slice = self.wait_for_slice(slice, progress=verbose, timeout=600)

        if create_slice:
            time.sleep(120)

        return_status, topology = self.slice_manager.get_slice_topology(slice_object=slice)
        if return_status != Status.OK:
            raise Exception("run_ssh_test failed to get topology. slice; {}, error {}".format(str(slice),str(topology)))

        #print("topology: {}".format(str(topology)))


        server_node = topology.nodes[test['server']['node_name']]
        server_name = server_node.name
        server_dataplane_ip='192.168.1.1'
        print("Config server: {}".format(str(server_name)))
        self.configure_test_node(server_node, server_dataplane_ip)

        node_num = 100
        nodes = []
        for node_name, node in topology.nodes.items():
            if node_name == server_name:
                continue

            dataplane_ip = '192.168.1.'+str(node_num)
            node_num = node_num + 1
            nodes.append({ 'dataplane_ip': dataplane_ip, 'node': node})
            print("Config node: {}".format(str(node_name)))
            self.configure_test_node(node, dataplane_ip)

        all_results = {}
        for node in nodes:
            result_name = "{}_{}_{}".format(test_name, server_name, node['node'].name)
            #self.run_tests(site1, site2, [self.latency_test])
            #self.run_tests(test_name, site1, site2, [self.latency_test, self.mtu_test, self.bandwidth_test, self.network_card_information, self.processor_information])
            #print("Run test: {}".format(str(server_node)))
            #print("Run test: {}".format(str(node['node'])))
            #print("Run test: {}".format(str(server_dataplane_ip)))
            #print("Run test: {}".format(str(node['dataplane_ip'])))
            #print("Run test: {} {} {} {}".format(str(server_node), str(node['node']), str(server_dataplane_ip), str(node['dataplane_ip'])))
            all_results[result_name] = self.run_tests(test_name, server_node, node['node'], server_dataplane_ip, node['dataplane_ip'], [self.latency_test, self.mtu_test, self.bandwidth_test])
            #all_results[result_name] = self.run_tests(test_name, server_node, node['node'], server_dataplane_ip, node['dataplane_ip'], [self.latency_test])

        return all_results

    def run(self, create_slice=True, run_test=True, delete=True):
        """
        Run the test
        :return:
        """



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author: Paul Ruth (pruth@renci.org)
from tabulate import tabulate
from ipaddress import IPv4Address

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from fabrictestbed_extensions.fablib.slice import Slice
    from fabrictestbed_extensions.fablib.node import Node
    from fabrictestbed_extensions.fablib.network_service import NetworkService
    from fabrictestbed_extensions.fablib.component import Component
    from fim.user.interface import Interface as FimInterface


class Interface:

    def __init__(self, component: Component = None, fim_interface: FimInterface = None):
        """
        Constructor. Sets keyword arguments as instance fields.

        :param component: the component to set on this interface
        :type component: Component
        :param fim_interface: the FABRIC information model interface to set on this fablib interface
        :type fim_interface: FimInterface
        """
        super().__init__()
        self.fim_interface = fim_interface
        self.component = component
        self.network = None

    def __str__(self):
        """
        Creates a tabulated string describing the properties of the interface.

        Intended for printing interface information.

        :return: Tabulated string of interface information
        :rtype: String
        """
        if self.get_network():
            network_name = self.get_network().get_name()
        else:
            network_name = None

        table = [   [ "Name", self.get_name() ],
                    [ "Network", network_name ],
                    [ "Bandwidth", self.get_bandwidth() ],
                    [ "VLAN", self.get_vlan() ],
                    [ "MAC", self.get_mac() ],
                    [ "Physical OS Interface", self.get_physical_os_interface_name() ],
                    [ "OS Interface", self.get_os_interface() ],
                    ]

        return tabulate(table)

    def get_os_interface(self) -> str:
        """
        Gets a name of the interface the operating system uses for this
        FABLib interface.

        If the interface requires a FABRIC VLAN tag, the interface name retruned
        will be the VLAN tagged.

        :return: OS interface name
        :rtype: String
        """
        try:
            #logging.debug(f"iface: {self}")
            os_iface = self.get_physical_os_interface_name()
            vlan = self.get_vlan()

            if vlan is not None:
                os_iface = f"{os_iface}.{vlan}"
        except:
            os_iface = None

        return os_iface

    def get_mac(self) -> str:
        """
        Gets the MAC addrress of the interface.

        :return: MAC address
        :rtype: String
        """
        try:
            os_iface = self.get_physical_os_interface()
            mac = os_iface['mac']
        except:
            mac = None

        return mac

    def get_physical_os_interface(self):
        """
        Not intended for API use
        """

        if self.get_network() is None:
            return None

        network_name = self.get_network().get_name()
        node_name = self.get_node().get_name()

        try:
            return self.get_slice().get_interface_map()[network_name][node_name]
        except:
            return None

    def get_physical_os_interface_name(self) -> str:
        """
        Gets a name of the physical interface the operating system uses for this
        FABLib interface.

        If the interface requires a FABRIC VLAN tag, the base interface name
        will be returned (i.e. not the VLAN tagged interface)

        :return: physicla OS interface name
        :rtype: String
        """
        if self.get_physical_os_interface():
            return self.get_physical_os_interface()['ifname']
        else:
            return None

    def config_vlan_iface(self):
        """
        Not intended for API use
        """
        if self.get_vlan() != None:
            self.get_node().add_vlan_os_interface(os_iface=self.get_physical_os_interface_name(),
                                                  vlan=self.get_vlan(), interface=self)

    def set_ip(self, ip=None, cidr=None, mtu=None):
        """
        Depricated
        """
        if cidr: cidr=str(cidr)
        if mtu: mtu=str(mtu)

        self.get_node().set_ip_os_interface(os_iface=self.get_physical_os_interface_name(),
                                            vlan=self.get_vlan(),
                                            ip=ip, cidr=cidr, mtu=mtu)

    def ip_addr_add(self, addr, subnet):
        """
        Add an IP address to the interface in the node.

        :param addr: IP address
        :type addr: IPv4Address or IPv6Address
        :param subnet: subnet
        :type subnet: IPv4Network or IPv4Network
        """
        self.get_node().ip_addr_add(addr, subnet, self)

    def ip_addr_del(self, addr, subnet):
        """
        Delete an IP address to the interface in the node.

        :param addr: IP address
        :type addr: IPv4Address or IPv6Address
        :param subnet: subnet
        :type subnet: IPv4Network or IPv4Network
        """
        self.get_node().ip_addr_del(addr, subnet, self)

    def ip_link_up(self):
        """
        Bring up the link on the interface.

        """
        self.get_node().ip_link_up(self)

    def ip_link_down(self):
        """
        Bring down the link on the interface.

        """
        self.get_node().ip_link_down(self)

    def set_vlan(self, vlan=None):
        """
        Set the VLAN on the FABRIC request.

        :param addr: vlan
        :type addr: String or int
        """
        if vlan: vlan=str(vlan)

        if_labels = self.get_fim_interface().get_property(pname="labels")
        if_labels.vlan = str(vlan)
        self.get_fim_interface().set_properties(labels=if_labels)

    def get_fim_interface(self) -> FimInterface:
        """
        Not intended for API use
        """
        return self.fim_interface

    def get_bandwidth(self) -> str:
        """
        Gets the bandwidth of an interface. Basic NICs claim 0 bandwidth but
        are 100 Gbps shared by all Basic NICs on the host.

        :return: bandwith
        :rtype: String
        """
        return self.get_fim_interface().capacities.bw

    def get_vlan(self) -> str:
        """
        Gets the FABRIC VLAN of an interface.

        :return: VLAN
        :rtype: String
        """
        try:
            vlan = self.get_fim_interface().get_property(pname="labels").vlan
        except:
            vlan = None
        return vlan

    def get_reservation_id(self) -> str or None:
        try:
            #TODO THIS DOESNT WORK.
            #print(f"{self.get_fim_interface()}")
            return self.get_fim_interface().get_property(pname='reservation_info').reservation_id
        except:
            return None

    def get_reservation_state(self) -> str or None:
        """
        Gets the reservation state

        :return: VLAN
        :rtype: String
        """
        try:
            return self.get_fim_interface().get_property(pname='reservation_info').reservation_state
        except:
            return None

    def get_error_message(self) -> str:
        """
        Gets the error messages

        :return: error
        :rtype: String
        """
        try:
            return self.get_fim_interface().get_property(pname='reservation_info').error_message
        except:
            return ""

    def get_name(self) -> str:
        """
        Gets the name of this interface.

        :return: the name of this interface
        :rtype: String
        """
        return self.get_fim_interface().name

    def get_component(self) -> Component:
        """
        Gets the component attached to this interface.

        :return: the component on this interface
        :rtype: Component
        """
        return self.component

    def get_model(self) -> str:
        """
        Gets the component model type on this interface's component.

        :return: the model of this interface's component
        :rtype: str
        """
        return self.get_component().get_model()

    def get_site(self) -> str:
        """
        Gets the site this interface's component is on.

        :return: the site this interface is on
        :rtype: str
        """
        return self.get_component().get_site()

    def get_slice(self) -> Slice:
        """
        Gets the FABLIB slice this interface's node is attached to.

        :return: the slice this interface is attached to
        :rtype: Slice
        """
        return self.get_node().get_slice()

    def get_node(self) -> Node:
        """
        Gets the node this interface's component is on.

        :return: the node this interface is attached to
        :rtype: Node
        """
        return self.get_component().get_node()

    def get_network(self) -> NetworkService:
        """
        Gets the network this interface is on.

        :return: the network service this interface is on
        :rtype: NetworkService
        """
        if hasattr(self, 'network'):
            #print(f"hasattr(self, 'network'): {hasattr(self, 'network')}, {self.network.get_name()}")
            return self.network
        else:
            for net in self.get_slice().get_networks():
                if net.has_interface(self):
                    self.network = net
                    #print(f"return found network, {self.network.get_name()}")
                    return self.network

        #print(f"hasattr(self, 'network'): {hasattr(self, 'network')}, None")
        return None



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author: Paul Ruth (pruth@renci.org)

from tabulate import tabulate

from fabrictestbed.slice_editor import Labels, Capacities

from fabrictestbed_extensions.fablib.slice import Slice


class FacilityPort:
    fim_interface = None
    slice = None

    def __init__(self, slice: Slice, fim_interface):
        """
        Constructor. Sets the fablib slice and FIM node based on arguments.
        :param slice: the fablib slice to have this node on
        :type slice: Slice
        :param node: the FIM node that this Node represents
        :type node: Node
        """
        super().__init__()
        self.fim_interface = fim_interface
        self.slice = slice

    def __str__(self):
        """
        Creates a tabulated string describing the properties of the node.
        Intended for printing node information.
        :return: Tabulated string of node information
        :rtype: String
        """
        table = [ ["name", self.get_name()], ]

        return tabulate(table) #, headers=["Property", "Value"])

    def get_fim_interface(self):
        return self.fim_interface

    def get_model(self):
        return "Facility_Port"

    def get_name(self):
        """
        Gets the name of the FABRIC node.
        :return: the name of the node
        :rtype: String
        """
        return self.get_fim_interface().name

    def get_site(self):
        return self.fim_interface.site

    @staticmethod
    def new_facility_port(slice=None, name=None, site=None, vlan=None, bandwidth=10):

        fim_facility_port = slice.get_fim_topology().add_facility(name=name, site=site,
                                                                 capacities=Capacities(bw=bandwidth),
                                                                 labels=Labels(vlan=vlan))
        return FacilityPort(slice, fim_facility_port)

    @staticmethod
    def get_facility_port(slice=None, facility_port=None):
        """
        Not intended for API call.
        Returns a new fablib node using existing FABRIC resources.
        :param slice: the fablib slice storing the existing node
        :type slice: Slice
        :param node: the FIM node stored in this fablib node
        :type node: Node
        :return: a new fablib node storing resources
        :rtype: Node
        """
        return FacilityPort(slice, facility_port)

    def get_slice(self):
        """
        Gets the fablib slice associated with this node.
        :return: the fablib slice on this node
        :rtype: Slice
        """
        return self.slice

    def get_name(self):

        try:
            return self.get_fim_node().name
        except:
            return None


    def get_interfaces(self):
        """
        Gets a particular interface associated with a FABRIC node.
        Accepts either the interface name or a network_name. If a network name
        is used this method will return the interface on the node that is
        connected to the network specified.
        If a name and network_name are both used, the interface name will
        take precedence.
        :param name: interface name to search for
        :type name: str
        :param network_name: network name to search for
        :type name: str
        :raise Exception: if interface is not found
        :return: an interface on the node
        :rtype: Interface
        """

        from fabrictestbed_extensions.fablib.interface import Interface

        ifaces = []
        for fim_interface in self.get_fim_interface().interface_list:
            ifaces.append(Interface(component=self, fim_interface=fim_interface))

        return ifaces




